# coding: utf-8

"""
    asset-lake-api-server

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from inspect import getfullargspec
from typing import TYPE_CHECKING, Any, List, Optional, Union

from pydantic.v1 import BaseModel, Field, StrictStr, ValidationError, validator

from clients.asset_lake.models.checkpoint_data import CheckpointData
from clients.asset_lake.models.file_data import FileData
from clients.asset_lake.models.lora_data import LoraData
from clients.asset_lake.models.textual_inversion_data import TextualInversionData
from clients.asset_lake.models.vae_data import VAEData
from clients.asset_lake.models.volume_data import VolumeData

DATA_ONE_OF_SCHEMAS = [
    "CheckpointData",
    "FileData",
    "LoraData",
    "TextualInversionData",
    "VAEData",
    "VolumeData",
]


class Data(BaseModel):
    """
    Data
    """

    # data type: FileData
    oneof_schema_1_validator: Optional[FileData] = None
    # data type: CheckpointData
    oneof_schema_2_validator: Optional[CheckpointData] = None
    # data type: LoraData
    oneof_schema_3_validator: Optional[LoraData] = None
    # data type: TextualInversionData
    oneof_schema_4_validator: Optional[TextualInversionData] = None
    # data type: VAEData
    oneof_schema_5_validator: Optional[VAEData] = None
    # data type: VolumeData
    oneof_schema_6_validator: Optional[VolumeData] = None
    if TYPE_CHECKING:
        actual_instance: Union[
            CheckpointData,
            FileData,
            LoraData,
            TextualInversionData,
            VAEData,
            VolumeData,
        ]
    else:
        actual_instance: Any
    one_of_schemas: List[str] = Field(DATA_ONE_OF_SCHEMAS, const=True)

    class Config:
        validate_assignment = True

    discriminator_value_class_map = {}

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                raise ValueError(
                    "If a position argument is used, only 1 is allowed to set `actual_instance`"
                )
            if kwargs:
                raise ValueError(
                    "If a position argument is used, keyword arguments cannot be used."
                )
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(**kwargs)

    @validator("actual_instance")
    def actual_instance_must_validate_oneof(cls, v):
        instance = Data.construct()
        error_messages = []
        match = 0
        # validate data type: FileData
        if not isinstance(v, FileData):
            error_messages.append(f"Error! Input type `{type(v)}` is not `FileData`")
        else:
            match += 1
        # validate data type: CheckpointData
        if not isinstance(v, CheckpointData):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `CheckpointData`"
            )
        else:
            match += 1
        # validate data type: LoraData
        if not isinstance(v, LoraData):
            error_messages.append(f"Error! Input type `{type(v)}` is not `LoraData`")
        else:
            match += 1
        # validate data type: TextualInversionData
        if not isinstance(v, TextualInversionData):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `TextualInversionData`"
            )
        else:
            match += 1
        # validate data type: VAEData
        if not isinstance(v, VAEData):
            error_messages.append(f"Error! Input type `{type(v)}` is not `VAEData`")
        else:
            match += 1
        # validate data type: VolumeData
        if not isinstance(v, VolumeData):
            error_messages.append(f"Error! Input type `{type(v)}` is not `VolumeData`")
        else:
            match += 1
        if match > 1:
            # more than 1 match
            raise ValueError(
                "Multiple matches found when setting `actual_instance` in Data with oneOf schemas: CheckpointData, FileData, LoraData, TextualInversionData, VAEData, VolumeData. Details: "
                + ", ".join(error_messages)
            )
        elif match == 0:
            # no match
            raise ValueError(
                "No match found when setting `actual_instance` in Data with oneOf schemas: CheckpointData, FileData, LoraData, TextualInversionData, VAEData, VolumeData. Details: "
                + ", ".join(error_messages)
            )
        else:
            return v

    @classmethod
    def from_dict(cls, obj: dict) -> Data:
        return cls.from_json(json.dumps(obj))

    @classmethod
    def from_json(cls, json_str: str) -> Data:
        """Returns the object represented by the json string"""
        instance = Data.construct()
        error_messages = []
        match = 0

        # use oneOf discriminator to lookup the data type
        _data_type = json.loads(json_str).get("asset_type")
        if not _data_type:
            raise ValueError(
                "Failed to lookup data type from the field `asset_type` in the input."
            )

        # check if data type is `CheckpointData`
        if _data_type == "CheckpointData":
            instance.actual_instance = CheckpointData.from_json(json_str)
            return instance

        # check if data type is `FileData`
        if _data_type == "FileData":
            instance.actual_instance = FileData.from_json(json_str)
            return instance

        # check if data type is `LoraData`
        if _data_type == "LoraData":
            instance.actual_instance = LoraData.from_json(json_str)
            return instance

        # check if data type is `TextualInversionData`
        if _data_type == "TextualInversionData":
            instance.actual_instance = TextualInversionData.from_json(json_str)
            return instance

        # check if data type is `VAEData`
        if _data_type == "VAEData":
            instance.actual_instance = VAEData.from_json(json_str)
            return instance

        # check if data type is `VolumeData`
        if _data_type == "VolumeData":
            instance.actual_instance = VolumeData.from_json(json_str)
            return instance

        # check if data type is `CheckpointData`
        if _data_type == "checkpoint":
            instance.actual_instance = CheckpointData.from_json(json_str)
            return instance

        # check if data type is `VAEData`
        if _data_type == "encoder":
            instance.actual_instance = VAEData.from_json(json_str)
            return instance

        # check if data type is `FileData`
        if _data_type == "file":
            instance.actual_instance = FileData.from_json(json_str)
            return instance

        # check if data type is `LoraData`
        if _data_type == "lora":
            instance.actual_instance = LoraData.from_json(json_str)
            return instance

        # check if data type is `TextualInversionData`
        if _data_type == "textual_inversion":
            instance.actual_instance = TextualInversionData.from_json(json_str)
            return instance

        # check if data type is `VolumeData`
        if _data_type == "volume":
            instance.actual_instance = VolumeData.from_json(json_str)
            return instance

        # deserialize data into FileData
        try:
            instance.actual_instance = FileData.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into CheckpointData
        try:
            instance.actual_instance = CheckpointData.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into LoraData
        try:
            instance.actual_instance = LoraData.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into TextualInversionData
        try:
            instance.actual_instance = TextualInversionData.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into VAEData
        try:
            instance.actual_instance = VAEData.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into VolumeData
        try:
            instance.actual_instance = VolumeData.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))

        if match > 1:
            # more than 1 match
            raise ValueError(
                "Multiple matches found when deserializing the JSON string into Data with oneOf schemas: CheckpointData, FileData, LoraData, TextualInversionData, VAEData, VolumeData. Details: "
                + ", ".join(error_messages)
            )
        elif match == 0:
            # no match
            raise ValueError(
                "No match found when deserializing the JSON string into Data with oneOf schemas: CheckpointData, FileData, LoraData, TextualInversionData, VAEData, VolumeData. Details: "
                + ", ".join(error_messages)
            )
        else:
            return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        to_json = getattr(self.actual_instance, "to_json", None)
        if callable(to_json):
            return self.actual_instance.to_json()
        else:
            return json.dumps(self.actual_instance)

    def to_dict(self) -> dict:
        """Returns the dict representation of the actual instance"""
        if self.actual_instance is None:
            return None

        to_dict = getattr(self.actual_instance, "to_dict", None)
        if callable(to_dict):
            return self.actual_instance.to_dict()
        else:
            # primitive type
            return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance"""
        return pprint.pformat(self.dict())



# coding: utf-8

"""
    FastAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import List, Union

from pydantic.v1 import BaseModel, Field, StrictFloat, StrictInt, conlist

from clients.image_gen.models.image_generation import ImageGeneration


class ImageGenerationResponse(BaseModel):
    """
    Response to an ImageGenerationRequest.  # noqa: E501
    """

    images: conlist(ImageGeneration) = Field(
        ..., description="List of ImageGeneration(s) generated by the request."
    )
    prediction_time_ms: Union[StrictFloat, StrictInt] = Field(
        ..., description="Total runtime of the image generation(s)."
    )
    seed: StrictInt = Field(
        ..., description="The seed used for the image generation(s)."
    )
    __properties = ["images", "prediction_time_ms", "seed"]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ImageGenerationResponse:
        """Create an instance of ImageGenerationResponse from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from pydantic.v1 by calling `to_dict()` of each item in images (list)
        _items = []
        if self.images:
            for _item in self.images:
                if _item:
                    _items.append(_item.to_dict())
            _dict["images"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> ImageGenerationResponse:
        """Create an instance of ImageGenerationResponse from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return ImageGenerationResponse.parse_obj(obj)

        _obj = ImageGenerationResponse.parse_obj(
            {
                "images": [
                    ImageGeneration.from_dict(_item) for _item in obj.get("images")
                ]
                if obj.get("images") is not None
                else None,
                "prediction_time_ms": obj.get("prediction_time_ms"),
                "seed": obj.get("seed"),
            }
        )
        return _obj



"""Generate Mustache templates that, when rendered, looks like a JSON structure."""


from typing import Optional, Type, cast

from pydantic.fields import FieldInfo as PydanticFieldInfo

from octoai.documenter.common import (
    Buffer,
    MustacheGenerationResult,
    MustacheTemplateGeneratingVisitor,
    MustacheTemplateRegistry,
    TypeFormatter,
)
from octoai.spec import BaseType
from octoai.spec.ast import (
    CustomObjectType,
    DictionaryType,
    EnumType,
    Field,
    FieldInfo,
    ListType,
    MediaKind,
    MediaType,
    OptionalType,
    PrimitiveType,
    PydanticHttpUrlType,
    SetType,
    SimpleFieldInfo,
    TupleType,
    UnionType,
    UnknownType,
)


class ConciseTypeFormatter(TypeFormatter):
    """Provide a concise format of the types which shows the type names only.

    Used when the ``JSONTypeFormatter`` chooses to format container types such as
    ``List``.
    """

    @classmethod
    def format_primitive(
        cls, primitive_type: PrimitiveType, field_info: FieldInfo
    ) -> str:
        """Show the name of the primitive type."""
        if primitive_type.type == bool:
            return "<true|false>"
        if primitive_type.type == str:
            return "<string>"
        elif primitive_type.type == bytes:
            return "<base64 string>"
        elif primitive_type.type == int:
            return "<integer>"
        elif primitive_type.type == float:
            return "<number>"
        else:
            raise ValueError(f"unsupported type {primitive_type.type.__name__}")

    @classmethod
    def format_union(cls, union_type: UnionType, field_info: FieldInfo) -> str:
        """Show the name of the union type."""
        return "<union>"

    @classmethod
    def format_list(cls, list_type: ListType, field_info: FieldInfo) -> str:
        """Show the name of the list type."""
        return "<list>"

    @classmethod
    def format_tuple(cls, tuple_type: TupleType, field_info: FieldInfo) -> str:
        """Show the name of the tuple type."""
        return "<tuple>"

    @classmethod
    def format_set(cls, set_type: SetType, field_info: FieldInfo) -> str:
        """Show the name of the set type."""
        return "<set>"

    @classmethod
    def format_dictionary(cls, dict_type: DictionaryType, field_info: FieldInfo) -> str:
        """Show the name of the dictionary type."""
        return "<dict>"

    @classmethod
    def format_enum(cls, enum_type: EnumType, field_info: FieldInfo) -> str:
        """Show the name of the enum type."""
        return f"<{enum_type.name}>"

    @classmethod
    def format_unknown(cls, _unknown_type: UnknownType, field_info: FieldInfo) -> str:
        """Show the name of the unknown type."""
        return "<unknown>"

    @classmethod
    def format_optional(cls, optional_type: OptionalType, field_info: FieldInfo) -> str:
        """Show the name of the optional type."""
        return "<optional>"

    @classmethod
    def format_pydantic_http_url(
        cls, _pydantic_http_url_type: PydanticHttpUrlType, field_info: FieldInfo
    ) -> str:
        """Show the name of the pydantic HttpUrl type."""
        return "<url string>"

    @classmethod
    def format_media(cls, media_type: MediaType, field_info: FieldInfo) -> str:
        """Show the name of the media type."""
        if media_type.kind == MediaKind.IMAGE:
            return "<image>"
        elif media_type.kind == MediaKind.AUDIO:
            return "<audio>"
        elif media_type.kind == MediaKind.VIDEO:
            return "<video>"
        else:
            raise ValueError(f"unsupported media type {media_type.kind}")

    @classmethod
    def format_custom_object(
        cls, custom_object_type: CustomObjectType, field_info: FieldInfo
    ) -> str:
        """Show the name of the custom object type."""
        return "<" + custom_object_type.name + ">"


class JSONTypeFormatter(TypeFormatter):
    """Provide a JSON-like display of the types.

    If examples are provided, the first example is returned instead.
    """

    @classmethod
    def _get_example_value(cls, field_info: FieldInfo) -> Optional[str]:
        if type(field_info) is SimpleFieldInfo or (
            type(field_info) is PydanticFieldInfo and not field_info.examples
        ):
            return None

        field_info = cast(PydanticFieldInfo, field_info)
        example = field_info.examples[0]
        return str(example)

    @classmethod
    def format_primitive(
        cls, primitive_type: PrimitiveType, field_info: FieldInfo
    ) -> str:
        """Show a JSON-like format for the primitive type."""
        example_value = cls._get_example_value(field_info)
        if example_value is not None:
            if primitive_type.type == str:
                return f'"{example_value}"'
            else:
                return str(example_value)

        if primitive_type.type == bool:
            return "<true|false>"
        if primitive_type.type == str:
            return "<string>"
        elif primitive_type.type == bytes:
            return "<base64 string>"
        elif primitive_type.type == int:
            return "<integer>"
        elif primitive_type.type == float:
            return "<number>"
        else:
            raise ValueError(f"unsupported type {primitive_type.type.__name__}")

    @classmethod
    def format_union(cls, union_type: UnionType, field_info: FieldInfo) -> str:
        """Format the union type."""
        return "|".join(
            ConciseTypeFormatter.format(member, field_info)
            for member in union_type.members
        )

    @classmethod
    def format_list(cls, list_type: ListType, field_info: FieldInfo) -> str:
        """Format the list type."""
        value = "[list of "
        value += ConciseTypeFormatter.format(list_type.member_type, field_info)
        value += "]"

        return value

    @classmethod
    def format_tuple(cls, tuple_type: TupleType, field_info: FieldInfo) -> str:
        """Format the tuple type."""
        value = "["
        value += ", ".join(
            ConciseTypeFormatter.format(member, field_info)
            for member in tuple_type.members
        )
        value += "]"

        return value

    @classmethod
    def format_set(cls, set_type: SetType, field_info: FieldInfo) -> str:
        """Format the set type."""
        value = "[set of "
        value += ConciseTypeFormatter.format(set_type.member_type, field_info)
        value += "]"

        return value

    @classmethod
    def format_dictionary(cls, dict_type: DictionaryType, field_info: FieldInfo) -> str:
        """Format the dictionary type."""
        value = "{"
        value += ConciseTypeFormatter.format(dict_type.key_type, field_info)
        value += ": "
        value += ConciseTypeFormatter.format(dict_type.value_type, field_info)
        value += "}"

        return value

    @classmethod
    def format_enum(cls, enum_type: EnumType, field_info: FieldInfo) -> str:
        """Format the enum type."""
        example_value = cls._get_example_value(field_info)
        if example_value is not None:
            return example_value

        formatted_variants = []
        for _, variant_value in enum_type.variants:
            if type(variant_value) is str:
                formatted_variants += [f'"{variant_value}"']
            else:
                formatted_variants += [str(variant_value)]
        return "|".join(formatted_variants)

    @classmethod
    def format_unknown(cls, _unknown_type: UnknownType, field_info: FieldInfo) -> str:
        """Format the unknown type."""
        return "<unknown>"

    @classmethod
    def format_optional(cls, optional_type: OptionalType, field_info: FieldInfo) -> str:
        """Format the optional type."""
        example_value = cls._get_example_value(field_info)
        if example_value is not None:
            return example_value

        value = ConciseTypeFormatter.format(optional_type.some_type, field_info)
        value += "|null"

        return value

    @classmethod
    def format_pydantic_http_url(
        cls, _pydantic_http_url_type: PydanticHttpUrlType, field_info: FieldInfo
    ) -> str:
        """Format the Pydantic HttpUrl type."""
        example_value = cls._get_example_value(field_info)
        if example_value:
            return f'"{example_value}"'

        return "<url string>"

    @classmethod
    def format_media(cls, media_type: MediaType, field_info: FieldInfo) -> str:
        """Format the media type."""
        example_value = cls._get_example_value(field_info)
        if example_value is not None:
            return example_value

        return '{"data": {"url": "<url string>"}}'

    @classmethod
    def format_custom_object(
        cls, custom_object_type: CustomObjectType, field_info: FieldInfo
    ) -> str:
        """Format the custom object type."""
        example_value = cls._get_example_value(field_info)
        if example_value is not None:
            return example_value

        return "<" + custom_object_type.name + ">"


class JSONGeneratingVisitor(MustacheTemplateGeneratingVisitor):
    """Generates a Mustache template that when rendered looks like a JSON structure."""

    _MAX_TRAVERSAL_DEPTH = 5

    def __init__(
        self,
        type_formatter_class: Type[TypeFormatter] = JSONTypeFormatter,
        base_buffer: Optional[Buffer] = None,
        template_registry: Optional[MustacheTemplateRegistry] = None,
    ):
        self._type_formatter_class = type_formatter_class
        super().__init__(base_buffer, template_registry)

    def generate(self, field: Field) -> MustacheGenerationResult:
        """Generate a Mustache template. This is the entry point.

        :param field: the field to generate a Mustache template for.
        """
        self.visit_base(field.type, field.info)
        return MustacheGenerationResult(
            generated_template=self._buffer.string,
            template_hash=self._template_registry.hash,
        )

    def visit_base(
        self,
        base_type: BaseType,
        field_info: FieldInfo,
    ):
        """Visit the base type.

        :param base_type: the base type.
        :param field_info: information about the field.
        """
        if type(base_type) is CustomObjectType:
            self.visit_custom_object(cast(CustomObjectType, base_type), field_info)
            return

        super().visit_base(base_type, field_info)

        value = self._type_formatter_class.format(base_type, field_info)
        self._template_registry.register_example(self._field_traversal_stack, value)
        self._buffer.print_string(
            self._template_registry.get_field_template_tag(self._field_traversal_stack)
        )

    def visit_custom_object(
        self,
        custom_object_type: CustomObjectType,
        field_info: FieldInfo,
    ):
        """Visit the custom object type.

        :param custom_object_type: The custom object type.
        :param field_info: information about the field.
        """
        # Add a space here to prevent colliding with Mustache opening brackets. This is
        # because we don't start a new line immediately when we see a field to avoid
        # display strangeness in the rendered string.
        self._buffer.print_string("{ ")

        # Filter out fields that will not be shown. This is needed to figure out when we
        # want to hide the commas. Note that this is purely to work around Mustache’s
        # logic-less limitations; we may consider switching to other methods in the
        # future.
        displayed_fields = {
            field_name: field_item
            for field_name, field_item in custom_object_type.fields.items()
            if self._should_display_field(field_item.info)
        }

        for idx, (field_name, field_item) in enumerate(displayed_fields.items()):
            self._field_traversal_stack.append(field_name)
            if len(self._field_traversal_stack) >= type(self)._MAX_TRAVERSAL_DEPTH:
                raise ValueError("Deeply nested objects are currently unsupported")

            if type(field_item.type) is not CustomObjectType:
                self._buffer.print_string(
                    self._template_registry.get_field_opening_format_tag(
                        self._field_traversal_stack
                    )
                )

            self._buffer.increase_indent()
            self._buffer.new_line()

            self._buffer.print_string('"' + field_name + '"' + ": ")

            self.visit_base(field_item.type, field_item.info)
            if idx < len(displayed_fields) - 1:
                self._buffer.print_string(",")

            if type(field_item.type) is not CustomObjectType:
                self._buffer.print_string(
                    self._template_registry.get_field_closing_format_tag(
                        self._field_traversal_stack
                    )
                )

            self._field_traversal_stack.pop()
            self._buffer.decrease_indent()

        self._buffer.new_line()
        self._buffer.print_string("}")



"""Example OctoAI service scaffold: Wav2Vec."""
import torch
from transformers import Wav2Vec2ForCTC, Wav2Vec2Processor

from octoai.service import Service
from octoai.types import Audio

"""
Wav2vec is a self-supervised learning algorithm for speech recognition.
"""


class Wav2VecService(Service):
    """An OctoAI service extends octoai.service.Service."""

    # Uncomment this function if you want to place large assets
    # (checkpoints, models, etc) on a docker volume instead of building
    # them into the image. This can provide faster
    # cold-start time. You may need to modify your inference pipeline
    # so that it leverages the locally-stored assets instead of
    # downloading them at container boot time.
    # def store_assets(self) -> None:
    #     """Download model assets."""
    #     # By default, instantiate the model to fill local disk
    #     # caches on the OctoAI volume. You can customize this if
    #     # your model can be fetched without instantiating it.
    #     self.setup()

    def setup(self):
        """Download model weights to disk."""
        self.processor = Wav2Vec2Processor.from_pretrained(
            "facebook/wav2vec2-base-960h"
        )
        self.model = Wav2Vec2ForCTC.from_pretrained("facebook/wav2vec2-base-960h")

    def infer(self, audio: Audio) -> str:
        """Perform inference with the model.

        The `Audio` type is a wrapper for binary audio files that provides a
        consistent API specification for your endpoint. The audio data is
        transferred over HTTP encoded as base64. This type provides support for
        encoding and decoding, reading audio data as numpy, creating audio files
        from numpy, and reading audio files from disk or remote URLs.

        See also the API reference at
        https://octoml.github.io/octoai-python-sdk/octoai.html#module-octoai.types
        """
        audio_array, sampling_rate = audio.to_numpy()
        input_values = self.processor(
            audio_array,
            sampling_rate=sampling_rate,
            return_tensors="pt",
            padding="longest",
        ).input_values
        logits = self.model(input_values).logits
        predicted_ids = torch.argmax(logits, dim=-1)
        transcription = self.processor.batch_decode(predicted_ids)

        return transcription[0]



# coding: utf-8

"""
    asset-lake-api-server

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, Dict, Optional

from pydantic.v1 import BaseModel, Field, StrictBool, StrictStr

from clients.asset_lake.models.asset_type import AssetType
from clients.asset_lake.models.transfer_api_type import TransferApiType


class CreateAssetRequest(BaseModel):
    """
    CreateAssetRequest
    """

    asset_type: AssetType = Field(...)
    data: Dict[str, Any] = Field(
        ...,
        description="Data specific to the asset type; use the class named ${asset_type}Data",
    )
    description: Optional[StrictStr] = Field(...)
    is_public: Optional[StrictBool] = False
    name: StrictStr = Field(...)
    transfer_api_type: Optional[TransferApiType] = None
    __properties = [
        "asset_type",
        "data",
        "description",
        "is_public",
        "name",
        "transfer_api_type",
    ]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> CreateAssetRequest:
        """Create an instance of CreateAssetRequest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # set to None if description (nullable) is None
        # and __fields_set__ contains the field
        if self.description is None and "description" in self.__fields_set__:
            _dict["description"] = None

        # set to None if transfer_api_type (nullable) is None
        # and __fields_set__ contains the field
        if (
            self.transfer_api_type is None
            and "transfer_api_type" in self.__fields_set__
        ):
            _dict["transfer_api_type"] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> CreateAssetRequest:
        """Create an instance of CreateAssetRequest from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return CreateAssetRequest.parse_obj(obj)

        _obj = CreateAssetRequest.parse_obj(
            {
                "asset_type": obj.get("asset_type"),
                "data": obj.get("data"),
                "description": obj.get("description"),
                "is_public": obj.get("is_public")
                if obj.get("is_public") is not None
                else False,
                "name": obj.get("name"),
                "transfer_api_type": obj.get("transfer_api_type"),
            }
        )
        return _obj



"""Example OctoAI service scaffold: Flan-T5-Small."""
from transformers import T5ForConditionalGeneration, T5Tokenizer

from octoai.service import Service

"""
Flan-T5 is an instruction-finetuned version of T5, a text-to-text
transformer language model.
"""


class T5Service(Service):
    """An OctoAI service extends octoai.service.Service."""

    # Uncomment this function if you want to place large assets
    # (checkpoints, models, etc) on a docker volume instead of building
    # them into the image. This can provide faster
    # cold-start time. You may need to modify your inference pipeline
    # so that it leverages the locally-stored assets instead of
    # downloading them at container boot time.
    # def store_assets(self) -> None:
    #     """Download model assets."""
    #     # By default, instantiate the model to fill local disk
    #     # caches on the OctoAI volume. You can customize this if
    #     # your model can be fetched without instantiating it.
    #     self.setup()

    def setup(self):
        """Download model weights to disk."""
        self.tokenizer = T5Tokenizer.from_pretrained("google/flan-t5-small")
        self.model = T5ForConditionalGeneration.from_pretrained("google/flan-t5-small")

    def infer(self, prompt: str) -> str:
        """Perform inference with the model."""
        inputs = self.tokenizer(prompt, return_tensors="pt")
        outputs = self.model.generate(**inputs)
        response = self.tokenizer.batch_decode(outputs, skip_special_tokens=True)

        return response[0]



from __future__ import annotations

# coding: utf-8

"""
    asset-lake-api-server

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import json
import pprint
import re  # noqa: F401
from enum import Enum


class BaseEngine(str, Enum):
    """
    BaseEngine
    """

    """
    allowed enum values
    """
    IMAGE_SLASH_STABLE_MINUS_DIFFUSION_MINUS_V1_MINUS_5 = "image/stable-diffusion-v1-5"
    IMAGE_SLASH_STABLE_MINUS_DIFFUSION_MINUS_XL_MINUS_V1_MINUS_0 = (
        "image/stable-diffusion-xl-v1-0"
    )
    TEXT_SLASH_LLAMA_MINUS_2_MINUS_7B = "text/llama-2-7b"

    @classmethod
    def from_json(cls, json_str: str) -> BaseEngine:
        """Create an instance of BaseEngine from a JSON string"""
        return BaseEngine(json.loads(json_str))



# coding: utf-8

"""
    asset-lake-api-server

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import io
import re  # noqa: F401
import warnings
from typing import Optional

from pydantic.v1 import (
    StrictBool,
    StrictInt,
    StrictStr,
    ValidationError,
    validate_arguments,
)

from clients.asset_lake.api_client import ApiClient
from clients.asset_lake.api_response import ApiResponse
from clients.asset_lake.exceptions import ApiTypeError, ApiValueError  # noqa: F401
from clients.asset_lake.models.asset_type import AssetType
from clients.asset_lake.models.base_engine import BaseEngine
from clients.asset_lake.models.complete_asset_upload_request import (
    CompleteAssetUploadRequest,
)
from clients.asset_lake.models.complete_asset_upload_response import (
    CompleteAssetUploadResponse,
)
from clients.asset_lake.models.create_asset_request import CreateAssetRequest
from clients.asset_lake.models.create_asset_response import CreateAssetResponse
from clients.asset_lake.models.data_type import DataType
from clients.asset_lake.models.delete_asset_response import DeleteAssetResponse
from clients.asset_lake.models.list_assets_response import ListAssetsResponse
from clients.asset_lake.models.retrieve_asset_response import RetrieveAssetResponse
from clients.asset_lake.models.transfer_api_type import TransferApiType


class AssetsApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client

    @validate_arguments
    def complete_asset_upload_v1_assets_asset_id_complete_upload_post(
        self,
        asset_id: StrictStr,
        complete_asset_upload_request: CompleteAssetUploadRequest,
        **kwargs
    ) -> CompleteAssetUploadResponse:  # noqa: E501
        """Complete Asset Upload  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.complete_asset_upload_v1_assets_asset_id_complete_upload_post(asset_id, complete_asset_upload_request, async_req=True)
        >>> result = thread.get()

        :param asset_id: (required)
        :type asset_id: str
        :param complete_asset_upload_request: (required)
        :type complete_asset_upload_request: CompleteAssetUploadRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: CompleteAssetUploadResponse
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the complete_asset_upload_v1_assets_asset_id_complete_upload_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.complete_asset_upload_v1_assets_asset_id_complete_upload_post_with_http_info(
            asset_id, complete_asset_upload_request, **kwargs
        )  # noqa: E501

    @validate_arguments
    def complete_asset_upload_v1_assets_asset_id_complete_upload_post_with_http_info(
        self,
        asset_id: StrictStr,
        complete_asset_upload_request: CompleteAssetUploadRequest,
        **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Complete Asset Upload  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.complete_asset_upload_v1_assets_asset_id_complete_upload_post_with_http_info(asset_id, complete_asset_upload_request, async_req=True)
        >>> result = thread.get()

        :param asset_id: (required)
        :type asset_id: str
        :param complete_asset_upload_request: (required)
        :type complete_asset_upload_request: CompleteAssetUploadRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(CompleteAssetUploadResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["asset_id", "complete_asset_upload_request"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method complete_asset_upload_v1_assets_asset_id_complete_upload_post"
                    % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}
        if _params["asset_id"]:
            _path_params["asset_id"] = _params["asset_id"]

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["complete_asset_upload_request"] is not None:
            _body_params = _params["complete_asset_upload_request"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = []  # noqa: E501

        _response_types_map = {
            "200": "CompleteAssetUploadResponse",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/v1/assets/{asset_id}/complete-upload",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def create_asset_v1_assets_post(
        self, create_asset_request: CreateAssetRequest, **kwargs
    ) -> CreateAssetResponse:  # noqa: E501
        """Create Asset  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.create_asset_v1_assets_post(create_asset_request, async_req=True)
        >>> result = thread.get()

        :param create_asset_request: (required)
        :type create_asset_request: CreateAssetRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: CreateAssetResponse
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the create_asset_v1_assets_post_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.create_asset_v1_assets_post_with_http_info(
            create_asset_request, **kwargs
        )  # noqa: E501

    @validate_arguments
    def create_asset_v1_assets_post_with_http_info(
        self, create_asset_request: CreateAssetRequest, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Create Asset  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.create_asset_v1_assets_post_with_http_info(create_asset_request, async_req=True)
        >>> result = thread.get()

        :param create_asset_request: (required)
        :type create_asset_request: CreateAssetRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(CreateAssetResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["create_asset_request"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method create_asset_v1_assets_post" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        if _params["create_asset_request"] is not None:
            _body_params = _params["create_asset_request"]

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["application/json"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = []  # noqa: E501

        _response_types_map = {
            "200": "CreateAssetResponse",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/v1/assets",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def delete_asset_v1_assets_asset_id_delete(
        self, asset_id: StrictStr, **kwargs
    ) -> DeleteAssetResponse:  # noqa: E501
        """Delete Asset  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.delete_asset_v1_assets_asset_id_delete(asset_id, async_req=True)
        >>> result = thread.get()

        :param asset_id: (required)
        :type asset_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: DeleteAssetResponse
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the delete_asset_v1_assets_asset_id_delete_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.delete_asset_v1_assets_asset_id_delete_with_http_info(
            asset_id, **kwargs
        )  # noqa: E501

    @validate_arguments
    def delete_asset_v1_assets_asset_id_delete_with_http_info(
        self, asset_id: StrictStr, **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Delete Asset  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.delete_asset_v1_assets_asset_id_delete_with_http_info(asset_id, async_req=True)
        >>> result = thread.get()

        :param asset_id: (required)
        :type asset_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(DeleteAssetResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["asset_id"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method delete_asset_v1_assets_asset_id_delete" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}
        if _params["asset_id"]:
            _path_params["asset_id"] = _params["asset_id"]

        # process the query parameters
        _query_params = []
        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # authentication setting
        _auth_settings = []  # noqa: E501

        _response_types_map = {
            "200": "DeleteAssetResponse",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/v1/assets/{asset_id}",
            "DELETE",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def list_assets_v1_assets_get(
        self,
        name: Optional[StrictStr] = None,
        is_public: Optional[StrictBool] = None,
        data_type: Optional[DataType] = None,
        asset_type: Optional[AssetType] = None,
        engine: Optional[BaseEngine] = None,
        limit: Optional[StrictInt] = None,
        offset: Optional[StrictInt] = None,
        owner: Optional[StrictStr] = None,
        asset_id: Optional[StrictStr] = None,
        **kwargs
    ) -> ListAssetsResponse:  # noqa: E501
        """List Assets  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_assets_v1_assets_get(name, is_public, data_type, asset_type, engine, limit, offset, owner, asset_id, async_req=True)
        >>> result = thread.get()

        :param name:
        :type name: str
        :param is_public:
        :type is_public: bool
        :param data_type:
        :type data_type: DataType
        :param asset_type:
        :type asset_type: AssetType
        :param engine:
        :type engine: BaseEngine
        :param limit:
        :type limit: int
        :param offset:
        :type offset: int
        :param owner:
        :type owner: str
        :param asset_id:
        :type asset_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: ListAssetsResponse
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the list_assets_v1_assets_get_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return self.list_assets_v1_assets_get_with_http_info(
            name,
            is_public,
            data_type,
            asset_type,
            engine,
            limit,
            offset,
            owner,
            asset_id,
            **kwargs
        )  # noqa: E501

    @validate_arguments
    def list_assets_v1_assets_get_with_http_info(
        self,
        name: Optional[StrictStr] = None,
        is_public: Optional[StrictBool] = None,
        data_type: Optional[DataType] = None,
        asset_type: Optional[AssetType] = None,
        engine: Optional[BaseEngine] = None,
        limit: Optional[StrictInt] = None,
        offset: Optional[StrictInt] = None,
        owner: Optional[StrictStr] = None,
        asset_id: Optional[StrictStr] = None,
        **kwargs
    ) -> ApiResponse:  # noqa: E501
        """List Assets  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_assets_v1_assets_get_with_http_info(name, is_public, data_type, asset_type, engine, limit, offset, owner, asset_id, async_req=True)
        >>> result = thread.get()

        :param name:
        :type name: str
        :param is_public:
        :type is_public: bool
        :param data_type:
        :type data_type: DataType
        :param asset_type:
        :type asset_type: AssetType
        :param engine:
        :type engine: BaseEngine
        :param limit:
        :type limit: int
        :param offset:
        :type offset: int
        :param owner:
        :type owner: str
        :param asset_id:
        :type asset_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(ListAssetsResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
            "name",
            "is_public",
            "data_type",
            "asset_type",
            "engine",
            "limit",
            "offset",
            "owner",
            "asset_id",
        ]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method list_assets_v1_assets_get" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []
        if _params.get("name") is not None:  # noqa: E501
            _query_params.append(("name", _params["name"]))

        if _params.get("is_public") is not None:  # noqa: E501
            _query_params.append(("is_public", _params["is_public"]))

        if _params.get("data_type") is not None:  # noqa: E501
            _query_params.append(("data_type", _params["data_type"].value))

        if _params.get("asset_type") is not None:  # noqa: E501
            _query_params.append(("asset_type", _params["asset_type"].value))

        if _params.get("engine") is not None:  # noqa: E501
            _query_params.append(("engine", _params["engine"].value))

        if _params.get("limit") is not None:  # noqa: E501
            _query_params.append(("limit", _params["limit"]))

        if _params.get("offset") is not None:  # noqa: E501
            _query_params.append(("offset", _params["offset"]))

        if _params.get("owner") is not None:  # noqa: E501
            _query_params.append(("owner", _params["owner"]))

        if _params.get("asset_id") is not None:  # noqa: E501
            _query_params.append(("asset_id", _params["asset_id"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # authentication setting
        _auth_settings = []  # noqa: E501

        _response_types_map = {
            "200": "ListAssetsResponse",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/v1/assets",
            "GET",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )

    @validate_arguments
    def retrieve_asset_v1_assets_asset_owner_and_name_or_id_get(
        self,
        asset_owner_and_name_or_id: StrictStr,
        transfer_api_type: Optional[TransferApiType] = None,
        **kwargs
    ) -> RetrieveAssetResponse:  # noqa: E501
        """Retrieve Asset  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.retrieve_asset_v1_assets_asset_owner_and_name_or_id_get(asset_owner_and_name_or_id, transfer_api_type, async_req=True)
        >>> result = thread.get()

        :param asset_owner_and_name_or_id: (required)
        :type asset_owner_and_name_or_id: str
        :param transfer_api_type:
        :type transfer_api_type: TransferApiType
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _request_timeout: timeout setting for this request.
               If one number provided, it will be total request
               timeout. It can also be a pair (tuple) of
               (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: RetrieveAssetResponse
        """
        kwargs["_return_http_data_only"] = True
        if "_preload_content" in kwargs:
            message = "Error! Please call the retrieve_asset_v1_assets_asset_owner_and_name_or_id_get_with_http_info method with `_preload_content` instead and obtain raw data from ApiResponse.raw_data"  # noqa: E501
            raise ValueError(message)
        return (
            self.retrieve_asset_v1_assets_asset_owner_and_name_or_id_get_with_http_info(
                asset_owner_and_name_or_id, transfer_api_type, **kwargs
            )
        )  # noqa: E501

    @validate_arguments
    def retrieve_asset_v1_assets_asset_owner_and_name_or_id_get_with_http_info(
        self,
        asset_owner_and_name_or_id: StrictStr,
        transfer_api_type: Optional[TransferApiType] = None,
        **kwargs
    ) -> ApiResponse:  # noqa: E501
        """Retrieve Asset  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.retrieve_asset_v1_assets_asset_owner_and_name_or_id_get_with_http_info(asset_owner_and_name_or_id, transfer_api_type, async_req=True)
        >>> result = thread.get()

        :param asset_owner_and_name_or_id: (required)
        :type asset_owner_and_name_or_id: str
        :param transfer_api_type:
        :type transfer_api_type: TransferApiType
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the ApiResponse.data will
                                 be set to none and raw_data will store the
                                 HTTP response body without reading/decoding.
                                 Default is True.
        :type _preload_content: bool, optional
        :param _return_http_data_only: response data instead of ApiResponse
                                       object with status code, headers, etc
        :type _return_http_data_only: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(RetrieveAssetResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = ["asset_owner_and_name_or_id", "transfer_api_type"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method retrieve_asset_v1_assets_asset_owner_and_name_or_id_get"
                    % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}
        if _params["asset_owner_and_name_or_id"]:
            _path_params["asset_owner_and_name_or_id"] = _params[
                "asset_owner_and_name_or_id"
            ]

        # process the query parameters
        _query_params = []
        if _params.get("transfer_api_type") is not None:  # noqa: E501
            _query_params.append(
                ("transfer_api_type", _params["transfer_api_type"].value)
            )

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))
        # process the form parameters
        _form_params = []
        _files = {}
        # process the body parameter
        _body_params = None
        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )  # noqa: E501

        # authentication setting
        _auth_settings = []  # noqa: E501

        _response_types_map = {
            "200": "RetrieveAssetResponse",
            "422": "HTTPValidationError",
        }

        return self.api_client.call_api(
            "/v1/assets/{asset_owner_and_name_or_id}",
            "GET",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),  # noqa: E501
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )



"""Generate cURL examples."""

from octoai.documenter.common import (
    Buffer,
    DocumentationGenerator,
    MustacheGenerationResult,
)
from octoai.documenter.json import JSONGeneratingVisitor
from octoai.spec.ast import Field, InferenceService


class CurlDocumentationGenerator(DocumentationGenerator):
    """Generate cURL documentation."""

    def generate(self, service: InferenceService) -> MustacheGenerationResult:
        """Generate usage example given the parsed ``InferenceService``.

        :param service: the parsed ``InferenceService``.
        """
        generation_template = self._get_generation_template()

        input_result = self._generate_input(service.inference_endpoint.inputs)

        full_result = generation_template.format(
            input_placeholder=input_result.generated_template,
        )

        return MustacheGenerationResult(
            generated_template=full_result,
            template_hash={
                "endpoint": "<endpoint>",
                "access_token": "$OCTOAI_TOKEN",
                **input_result.template_hash,
            },
        )

    def _get_generation_template(self) -> str:
        return """curl {{{{{{endpoint}}}}}}/infer \\
    -X POST \\{{{{#access_token}}}}
    -H "Authorization: Bearer {{{{{{access_token}}}}}}" \\{{{{/access_token}}}}
    -d @- << 'EOF'{input_placeholder}
    EOF
"""

    def _generate_input(self, field: Field) -> MustacheGenerationResult:
        generator = JSONGeneratingVisitor(base_buffer=Buffer(indentation_level=1))
        data = generator.generate(field)
        return data



"""Example OctoAI service scaffold: YOLOv8."""
import json
from typing import List

import numpy as np
from PIL import Image as PImage
from pydantic import BaseModel, Field
from ultralytics import YOLO

from octoai.service import Service
from octoai.types import Image

"""
YOLO (You Only Look Once) is a real-time object detection algorithm.
YOLO works by dividing an image into a grid of cells. Each cell predicts
the probability of each object class being present in the cell, as well
as the bounding box coordinates of the object.

The `YOLOv8Service.infer()` method returns a `YOLOResponse`, which represents
a list of `Detections`. Each `Detection` represents a bounding box with a
class label and confidence value. A `YOLOResponse` also contains an `Image`
object, which represets a serialized binary image file encoded as base64
that has the detection boxes drawn in it. See also the comments inside
`YOLOv8Service.infer()` below.

The YOLOv8 model returns a list of detections that look like this:

    [{
        'name': 'bus',
        'class': 5,
        'confidence': 0.95,
        'box': {
            'x1': 2.91,
            'x2': 809.51,
            'y1': 230.68,
            'y2': 881.00
        }
    }, ...]

For more information about defining your own entities, such as `YOLOResponse`,
`Detection` and `Box` as in this example, see:
https://docs.octoai.cloud/docs/containerize-your-code-with-our-cli#appendix-openapi-specification-and-pydantic-types

"""


class Box(BaseModel):
    """Represents corners of a detection box."""

    x1: float
    x2: float
    y1: float
    y2: float


class Detection(BaseModel):
    """Represents a detection."""

    name: str
    class_: int = Field(..., alias="class")
    confidence: float
    box: Box


class YOLOResponse(BaseModel):
    """Response includes list of detections and rendered image."""

    detections: List[Detection]
    image_with_boxes: Image


class YOLOv8Service(Service):
    """An OctoAI service extends octoai.service.Service."""

    # Uncomment this function if you want to place large assets
    # (checkpoints, models, etc) on a docker volume instead of building
    # them into the image. This can provide faster
    # cold-start time. You may need to modify your inference pipeline
    # so that it leverages the locally-stored assets instead of
    # downloading them at container boot time.
    # def store_assets(self) -> None:
    #     """Download model assets."""
    #     # By default, instantiate the model to fill local disk
    #     # caches on the OctoAI volume. You can customize this if
    #     # your model can be fetched without instantiating it.
    #     self.setup()

    def setup(self):
        """Download model weights to disk."""
        self.model = YOLO("yolov8l.pt")

    def infer(self, image: Image) -> YOLOResponse:
        """Perform inference with the model.

        The `Image` type is a wrapper for binary image files that provides a
        consistent API specification for your endpoint. The image data is
        transferred over HTTP encoded as base64. This type provides support for
        encoding and decoding, reading image data as Pillow, creating image files
        from Pillow, and reading image files from disk or remote URLs.

        See also the API reference at
        https://octoml.github.io/octoai-python-sdk/octoai.html#module-octoai.types

        The `YOLOResponse` class is transferred over HTTP as JSON as described
        above.
        """
        image_pil = image.to_pil()
        output = self.model(image_pil)

        # no Python dict available directly from the model
        detections = json.loads(output[0].tojson())

        # get the rendered image with bounding boxes and labels
        # - Result.plot() generates BGR; flip to RGB
        img_out_numpy = np.flip(output[0].plot(), axis=2)
        img_out_pil = PImage.fromarray(img_out_numpy)

        # Return detection data and a rendered image with boxes
        return YOLOResponse(
            detections=[Detection(**d) for d in detections],
            image_with_boxes=Image.from_pil(img_out_pil),
        )



from __future__ import annotations

# coding: utf-8

"""
    FastAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import json
import pprint
import re  # noqa: F401
from enum import Enum


class Scheduler(str, Enum):
    """
    The schedulers available for image generation.
    """

    """
    allowed enum values
    """
    PNDM = "PNDM"
    LMS = "LMS"
    KLMS = "KLMS"
    DDIM = "DDIM"
    DDPM = "DDPM"
    HEUN = "HEUN"
    K_HEUN = "K_HEUN"
    K_EULER = "K_EULER"
    K_EULER_ANCESTRAL = "K_EULER_ANCESTRAL"
    DPM_SOLVER_MULTISTEP = "DPM_SOLVER_MULTISTEP"
    DPM_PLUS_PLUS_2_M_KARRAS = "DPM_PLUS_PLUS_2M_KARRAS"
    DPM_SINGLE = "DPM_SINGLE"
    DPM_2 = "DPM_2"
    DPM_2_ANCESTRAL = "DPM_2_ANCESTRAL"
    DPM_PLUS_PLUS_SDE_KARRAS = "DPM_PLUS_PLUS_SDE_KARRAS"
    UNI_PC = "UNI_PC"

    @classmethod
    def from_json(cls, json_str: str) -> Scheduler:
        """Create an instance of Scheduler from a JSON string"""
        return Scheduler(json.loads(json_str))



"""Asset Orchestrator class."""
from __future__ import (
    annotations,  # required to allow 3.7+ python use type | syntax introduced in 3.10
)

import logging
import os
import re
from abc import ABC
from pathlib import Path
from typing import List, Union

import boto3
import httpx
import yaml

import clients.asset_lake.models as asset_models
import octoai  # for version header
from clients.asset_lake import ApiClient, AssetsApi, Configuration
from clients.asset_lake.models import Status  # used in docstring
from clients.asset_lake.models import (
    AssetType,
    BaseEngine,
    CreateAssetRequest,
    DataType,
    DeleteAssetResponse,
    FileExtension,
    FileFormat,
    PresignedUrlTransferApi,
    StsTransferApi,
    TransferApiType,
)
from octoai.errors import OctoAIServerError, OctoAIValidationError

LOG = logging.getLogger(__name__)


class Asset:
    """Asset Orchestrator implementation of an asset.

    Hashable with UUID used as key if passed to a dictionary.
    :param id: UUID unique to asset.
    :param asset_type: :class:`AssetType`, including "lora", "checkpoint", or
        "textual-inversion".
    :param name: Alphanumeric, _, or - allowed.
    :param description: Description of asset.
    :param size_bytes: Total bytes of the asset file.
    :param status: :class:`Status` of Asset.  One of "ready_to_upload", "ready",
        "uploaded", "deleted", "rejected", or "error".
    :param status_details: Description of asset status.
    :param created_at: Time created.
    :param data: Additional information about asset such as engine, file_format, etc.
    :param tenant_uuid: UUID of person who created the asset.
    """

    def __init__(
        self,
        id,
        asset_type,
        name,
        description,
        size_bytes,
        status,
        status_details,
        created_at,
        data,
        tenant_uuid,
        *args,
        **kwargs,  # Catch any vars from al.asset type we don't care about
    ):
        self.id = id
        self.asset_type = asset_type
        self.name = name
        self.description = description
        self.size_bytes = size_bytes
        self.status = status
        self.status_details = status_details
        self.created_at = created_at
        self.data = data
        self.tenant_uuid = tenant_uuid

    def __hash__(self):
        return hash(self.id)

    def __str__(self):
        return f"id: {self.id}, name: {self.name}, status: {self.status}\n"


class FileData:
    def __init__(self, file_format: FileExtension | str, version: str = ""):
        try:
            file_format = FileExtension(file_format)
        except ValueError:
            msg = (
                f"file_format({file_format}) in `data` field is not valid.  "
                f"Valid options include 'png', 'jpg', 'jpeg', and 'txt'."
            )
            raise OctoAIValidationError(msg)

        self._asset_model = asset_models.FileData(
            asset_type="file", file_format=file_format, version=version
        )

        self.asset_type = AssetType.FILE
        self.file_format = file_format
        self.version = version


class ModelData(ABC):
    """
    Base class for Checkpoints, LoRAs, and Textual Inversions.

    :param engine: Compatible :class:`BaseEngine` type for model.  Includes
        "image/stable-diffusion-v1-5" or "image/stable-diffusion-xl-v1-0".
    :param file_format: :class:`FileFormat` of model, includes `safetensors`.
    :param data_type: :class:`DataType` or str matching an enum in DataType, default to
        'fp16'.
    """

    def __init__(
        self,
        engine: BaseEngine | str,
        file_format: FileFormat | str,
        data_type: DataType | str,
    ):
        try:
            if type(file_format) == str:
                file_format = FileFormat(file_format)
        except ValueError:
            msg = (
                f"file_format({file_format}) in `data` field is not valid.  "
                f"Valid options include 'safetensors'."
            )
            raise OctoAIValidationError(msg)
        if isinstance(file_format, FileFormat):  # Linter handling for unions
            self.file_format = file_format

        try:
            if type(engine) == str:
                engine = BaseEngine(engine)
        except ValueError:
            msg = (
                f"engine({engine}) in `data` field is not valid. Valid options include "
                f"'image/stable-diffusion-v1-5' or 'image/stable-diffusion-xl-v1-0'."
            )
            raise OctoAIValidationError(msg)
        if isinstance(engine, BaseEngine):
            self.engine = engine

        try:
            if type(data_type) == str:
                data_type = DataType(data_type)
        except ValueError:
            msg = (
                f"data_type({data_type}) in `data` field is not valid.  "
                f"Valid options include 'fp16' and 'fp32'."
            )
            raise OctoAIValidationError(msg)
        if isinstance(data_type, DataType):
            self.data_type = data_type


class CheckpointData(ModelData):
    """
    Checkpoint data associated with `checkpoint` AssetType.

    Used for :meth:`AssetOrchestrator.create`.
    :param engine: Compatible :class:`BaseEngine` type for model.  Includes
        "image/stable-diffusion-v1-5" or "image/stable-diffusion-xl-v1-0".
    :param file_format: :class:`FileFormat` of model, includes `safetensors`.
    :param data_type: :class:`DataType` or str matching an enum in DataType, default to
        'fp16'.
    """

    def __init__(
        self,
        engine: BaseEngine | str,
        file_format: FileFormat | str,
        data_type: DataType | str = DataType.FP16,
    ):
        # Validates and sets params
        super(CheckpointData, self).__init__(engine, file_format, data_type)
        self._asset_model = asset_models.CheckpointData(
            data_type=self.data_type,
            engine=self.engine,
            file_format=self.file_format,
            asset_type=AssetType.CHECKPOINT,
        )
        self.asset_type = AssetType.CHECKPOINT


class LoraData(ModelData):
    """LoRA data associated with the `lora` AssetType.

    Used for :meth:`AssetOrchestrator.create`.
    :param engine: Compatible :class:`BaseEngine` type for model.  Includes
        "image/stable-diffusion-v1-5" or "image/stable-diffusion-xl-v1-0".
    :param file_format: :class:`FileFormat` of model, includes `safetensors`.
    :param data_type: :class:`DataType` or str matching an enum in DataType, default to
        'fp16'.
    """

    def __init__(
        self,
        engine: BaseEngine | str,
        file_format: FileFormat | str,
        data_type: DataType | str = DataType.FP16,
    ):
        super(LoraData, self).__init__(engine, file_format, data_type)
        self._asset_model = asset_models.LoraData(
            data_type=self.data_type,
            engine=self.engine,
            file_format=self.file_format,
            asset_type=AssetType.LORA,
        )
        self.asset_type = AssetType.LORA


class TextualInversionData(ModelData):
    """TextualInversionData associated with `textual_inversion` :class:`AssetType`.

    Used for :meth:`AssetOrchestrator.create`.
    :param engine: Compatible :class:`BaseEngine` type for model.  Includes
        "image/stable-diffusion-v1-5" or "image/stable-diffusion-xl-v1-0".
    :param file_format: :class:`FileFormat` of model, includes `safetensors`.
    :param data_type: :class:`DataType` or str matching an enum in DataType, default to
        'fp16'.
    """

    def __init__(
        self,
        engine: BaseEngine | str,
        file_format: FileFormat | str,
        trigger_words: List[str],
        data_type: DataType | str = DataType.FP16,
    ):
        super(TextualInversionData, self).__init__(engine, file_format, data_type)
        self._asset_model = asset_models.TextualInversionData(
            data_type=self.data_type,
            engine=self.engine,
            file_format=self.file_format,
            asset_type=AssetType.TEXTUAL_INVERSION,
            trigger_words=trigger_words,
        )
        self.asset_type = AssetType.TEXTUAL_INVERSION
        self.trigger_words = trigger_words


AssetData = Union[
    FileData,
    CheckpointData,
    LoraData,
    TextualInversionData,
]


class AssetOrchestrator:
    """Asset Orchestrator class to create, read, delete, and list assets.

    :param token: OCTOAI_TOKEN if one is not set as an environment variable, default to
        None.
    :param config_path: Path to config file from CLI, used if token or envvar is not
        set, default to None and checks default path.
    :param asset_orch_api_endpoint: Defaults to "https://api.octoai.cloud/".
    """

    def __init__(
        self,
        token: str | None = None,
        config_path: str | None = None,
        asset_orch_api_endpoint: str = None,
    ):
        # Would benefit from refactoring this with client code
        if token is None:
            token = token if token else os.environ.get("OCTOAI_TOKEN", None)

        if not token:
            # Default path is ~/.octoai/config.yaml for token, can be overridden
            path = Path(config_path) if config_path else Path.home()
            try:
                with open(
                    (path / Path(".octoai/config.yaml")), encoding="utf-8"
                ) as octoai_config_yaml:
                    config_dict = yaml.safe_load(octoai_config_yaml)
                token = config_dict.get("token")
            except FileNotFoundError:
                token = None

        if not token:
            msg = (
                "Authorization is required.  Please set an `OCTOAI_TOKEN` "
                "environment variable, or pass your token to the client using "
                "`asset_client = AssetOrchestrator(token='your-octoai-api-token')`"
            )
            raise OctoAIValidationError(msg)
        if asset_orch_api_endpoint is None:
            asset_orch_api_endpoint = "https://api.octoai.cloud/"
        conf = Configuration(host=asset_orch_api_endpoint)
        temp_client = ApiClient(
            conf, header_name="Authorization", header_value=f"Bearer {token}"
        )
        version = octoai.__version__
        temp_client.user_agent = f"octoai-{version}"
        self._auth_header = f"Bearer {token}"
        self.client = AssetsApi(api_client=temp_client)

    def create(
        self,
        file: str,
        data: AssetData,
        name: str,  # only alphanumeric characters, _, -
        description: str | None = None,
        transfer_api_type: str | TransferApiType | None = None,
    ) -> Asset:
        """
        Create and upload an asset.

        :param file: str to file path
        :param data: :class:`CheckpointData`, :class:`LoraData`,
            or :class:`TextualInversionData`
        :param name: Name of asset, alphanumeric with `-` and `_` characters allowed.
        :param description: Description of asset, optional, defaults to `None`.
        :param transfer_api_type: :class:`TransferApiType` or str of either
            "presigned-url", or "sts", defaults to "sts" for >= 50mb and
            "presigned-url" for under 50mb.
        :return: :class:`Asset`
        """
        self._validate_name(name)
        if transfer_api_type is None:  # Chooses best option for file size
            if os.path.getsize(file) >= 52428800:
                transfer_api_type = TransferApiType.STS
            else:
                transfer_api_type = TransferApiType.PRESIGNED_MINUS_URL
        inputs = locals()  # Clean up vars for create asset request
        inputs.pop("self")
        inputs.pop("file")
        inputs["asset_type"] = data.asset_type
        # Probably should make this a parse obj to dict vs storing ref to al model
        inputs["data"] = data._asset_model.to_dict()

        request = CreateAssetRequest.from_dict(inputs)

        response = self.client.create_asset_v1_assets_post(request)
        x_api = response.transfer_api.actual_instance
        if isinstance(x_api, PresignedUrlTransferApi):
            with open(file, "rb") as file_data:
                upload_resp = httpx.put(
                    url=x_api.put_url, content=file_data, timeout=60000
                )
        if isinstance(x_api, StsTransferApi):
            self._sts_upload(x_api, file)

        # Get asset after upload to verify correct status
        self.client.complete_asset_upload_v1_assets_asset_id_complete_upload_post(
            response.asset.id,
            asset_models.CompleteAssetUploadRequest(token=response.token),
        )
        asset = self.get(response.asset.id)
        return asset

    @staticmethod
    def _sts_upload(x_api: StsTransferApi, file: str):
        s3_client = boto3.client(
            "s3",
            aws_access_key_id=x_api.aws_access_key_id,
            aws_secret_access_key=x_api.aws_secret_access_key,
            aws_session_token=x_api.aws_session_token,
        )
        try:
            s3_client.upload_file(file, x_api.s3_bucket, x_api.s3_key)
        except Exception as e:
            raise OctoAIServerError(f"Error uploading file to server: {e}")

    def list(
        self,
        name: str | None = None,
        data_type: DataType | None = None,
        data: AssetType | dict | None = None,
        engine: BaseEngine | None = None,
        limit: int | None = None,
        offset: int | None = None,
        owner: str | None = None,
    ) -> list[Asset]:
        """
        Return list of assets filtered on the non-None parameters.

        :param name: Asset name, alphanumeric, -, and _ allowed.  Defaults to None.
        :param data_type: :class:`DataType`, defaults to None.
        :param data: :class:`AssetType` or dict, defaults to None.
        :param engine: :class:`BaseEngine` of asset, defaults to None.
        :param limit: Max number of assets to return, defaults to None.
        :param offset: Where to start including next list of assets, defaults to None.
        :param owner: ID of owner, defaults to None.
        :return: list[:class:`Asset`]
        """
        if name is not None:
            self._validate_name(name)
        inputs = locals()
        inputs.pop("self")
        for key in list(inputs):
            if inputs[key] is None:
                inputs.pop(key)
        response = self.client.list_assets_v1_assets_get(**inputs)
        result = []
        for item in response.data:
            result.append(Asset(**item.to_dict()))
        return result

    def get(self, asset_name_or_id: str) -> Asset:
        """
        Get an asset associated with an asset_id.

        :param asset_name_or_id: ID or name of asset to get.
        :return: :class:`Asset`
        """
        self._validate_name(asset_name_or_id)
        response = self.client.retrieve_asset_v1_assets_asset_owner_and_name_or_id_get(
            asset_name_or_id
        )
        asset = Asset(**response.asset.to_dict())
        return asset

    def delete(self, asset_id: str | None) -> DeleteAssetResponse:
        """Delete an asset.

        :param asset_id: the UUID of the asset to be deleted.
        :return: :class:`DeleteAssetResponse` containing an `id` as a str and
            `deleted_at` as a str with the timestamp.
        """
        response = self.client.delete_asset_v1_assets_asset_id_delete(asset_id)
        return response

    @staticmethod
    def _validate_name(name: str):
        if re.fullmatch("^[a-zA-Z0-9_-]*$", name) is None:
            msg = (
                f"name({name}) is invalid.  Valid names may only contain "
                f"alphanumeric, '-', or '_' characters."
            )
            raise OctoAIValidationError(msg)



# coding: utf-8

"""
    asset-lake-api-server

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional

from pydantic.v1 import BaseModel, Field, StrictBool, StrictInt, StrictStr

from clients.asset_lake.models.asset_type import AssetType
from clients.asset_lake.models.data import Data
from clients.asset_lake.models.status import Status


class Asset(BaseModel):
    """
    Asset
    """

    asset_type: AssetType = Field(...)
    created_at: StrictStr = Field(...)
    data: Data = Field(...)
    description: StrictStr = Field(...)
    id: StrictStr = Field(...)
    is_public: Optional[StrictBool] = False
    name: StrictStr = Field(...)
    size_bytes: StrictInt = Field(...)
    status: Status = Field(...)
    status_details: StrictStr = Field(...)
    tenant_uuid: StrictStr = Field(...)
    __properties = [
        "asset_type",
        "created_at",
        "data",
        "description",
        "id",
        "is_public",
        "name",
        "size_bytes",
        "status",
        "status_details",
        "tenant_uuid",
    ]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Asset:
        """Create an instance of Asset from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from pydantic.v1 by calling `to_dict()` of data
        if self.data:
            _dict["data"] = self.data.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> Asset:
        """Create an instance of Asset from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return Asset.parse_obj(obj)

        _obj = Asset.parse_obj(
            {
                "asset_type": obj.get("asset_type"),
                "created_at": obj.get("created_at"),
                "data": Data.from_dict(obj.get("data"))
                if obj.get("data") is not None
                else None,
                "description": obj.get("description"),
                "id": obj.get("id"),
                "is_public": obj.get("is_public")
                if obj.get("is_public") is not None
                else False,
                "name": obj.get("name"),
                "size_bytes": obj.get("size_bytes"),
                "status": obj.get("status"),
                "status_details": obj.get("status_details"),
                "tenant_uuid": obj.get("tenant_uuid"),
            }
        )
        return _obj



"""
Server for OctoAI endpoints created with the ``octoai`` CLI.

Developers that want to create an endpoint should not use
this module directly. Instead, they should use the ``octoai``
command-line interface, which directs them to implement the
``octoai.service.Service`` class and use the ``octoai`` CLI to help
build and deploy their endpoint.
"""
import asyncio
import dataclasses
import enum
import importlib
import inspect
import json
import logging
import multiprocessing
import os
import signal
import sys
import time
from contextlib import asynccontextmanager
from enum import Enum
from http import HTTPStatus
from multiprocessing import Pipe, Process, Queue
from multiprocessing.connection import Connection
from typing import Any, Dict, NamedTuple, Optional, Type

import chevron
import click
import uvicorn
from fastapi import FastAPI
from fastapi.responses import JSONResponse, Response
from pydantic import BaseModel

from octoai import documenter
from octoai.documenter import MustacheGenerationResult

from .service import (
    STORE_ASSETS_NOT_OVERRIDDEN,
    VOLUME_ENVIRONMENT,
    ResponseAnalytics,
    Service,
    implements_form_data,
    inspect_input_types,
    inspect_output_types,
    transform_form_data_signature,
)

_LOG = logging.getLogger(__name__)

_OCTOAI_SERVICE_MODULE = "octoai.service"
_OCTOAI_BASE_SERVICE_CLASS = "Service"

_PREDICT_LOOP_WATCHDOG_SECONDS = 2
"""Delay in seconds between checking if the predict loop is running."""

_process_mutex = multiprocessing.Lock()
"""Lock for spawning predict loop."""

_TERMINATE_PREDICT_LOOP_REQUEST = "terminate"
"""Sent to predict loop process to terminate it."""


_LOAD_MODEL_MESSAGE = "load_model"
"""Sent from predict loop when it is loading the model."""


_MODEL_LOADED_MESSAGE = "model_loaded"
"""Sent from predict loop when it is ready for inference."""

_STORE_ASSET_NEEDED_CODE = 10
"""Exit status code when store asset is required."""

_STORE_ASSET_NOT_NEEDED_CODE = 11
"""Exit status code when store asset is not required."""


class ServiceMethod(Enum):
    """Class for distinguishing route implementations in the request queue."""

    INFER_JSON = "infer"
    INFER_FORM_DATA = "infer_form_data"


class InferenceRequest(NamedTuple):
    """Class for returning inference results."""

    response_pipe: Connection
    method: ServiceMethod
    inputs: Any


class InferenceResponse(NamedTuple):
    """Class for returning inference results."""

    inference_time_ms: float
    outputs: Any


def is_store_assets_needed(ctx) -> bool:
    """Check if store-assets step is required."""

    def check_store_assets_overriden(queue, ctx, dummy) -> None:
        service = load_service(
            ctx.parent.params["service_module"],
            class_name=ctx.parent.params["service_class"],
        )
        queue.put(not hasattr(service.store_assets, STORE_ASSETS_NOT_OVERRIDDEN))

    queue: Queue[bool] = Queue()
    p = Process(target=check_store_assets_overriden, args=(queue, ctx, 0))
    p.start()
    p.join(timeout=10)
    result = queue.get()
    return result


def maybe_set_volume_environment_variables(ctx) -> None:
    """Set volume environment variables if needed."""
    if not is_store_assets_needed(ctx):
        return
    for key, value in VOLUME_ENVIRONMENT.items():
        os.environ[key] = value


def _predict_loop(service, _request_queue):
    """Loop which handles prediction requests.

    This loop runs for the duration of the server and receives prediction
    requests posted to the _REQUEST_QUEUE. When the request is done processing
    the results are posted to the response_pipe where they are handled by
    the main /predict endpoint.
    """
    startup_pipe = _request_queue.get()
    startup_pipe.send(_LOAD_MODEL_MESSAGE)
    try:
        service.setup()
    except Exception as e:
        _LOG.error("_predict_loop: model setup failed with {e}", exc_info=1)
        startup_pipe.send(e)
        startup_pipe.close()
        return

    startup_pipe.send(_MODEL_LOADED_MESSAGE)
    startup_pipe.close()

    def signal_handler(_signum, _frame):
        # This will only kill the _predict_loop process, not the parent
        sys.exit()

    signal.signal(signal.SIGINT, signal_handler)

    while True:
        try:
            inference_request = _request_queue.get()
            if (
                isinstance(inference_request, type(_TERMINATE_PREDICT_LOOP_REQUEST))
                and inference_request == _TERMINATE_PREDICT_LOOP_REQUEST
            ):
                sys.exit()

            try:
                start_time = time.perf_counter_ns()
                infer_fn = getattr(service, inference_request.method.value)
                results = infer_fn(**inference_request.inputs)
                stop_time = time.perf_counter_ns()
                response = InferenceResponse((stop_time - start_time) / 1e9, results)
            except Exception as e:
                _LOG.error("infer() raised Exception", exc_info=1)
                response = e
            inference_request.response_pipe.send(response)
            inference_request.response_pipe.close()
        except Exception:
            # We only end up here if something went wrong outside the predict call
            # continue loop
            pass


class Server:
    """
    Server for OctoAI endpoints created with the ``octoai`` CLI.

    Developers that want to create an endpoint should not use
    this class directly. Instead, they should use the ``octoai``
    command-line interface, which directs them to implement the
    ``octoai.service.Service`` class and use the ``octoai`` CLI to
    help build and deploy their endpoint.
    """

    class State(enum.Enum):
        """Describes the states of Server."""

        UNINITIALIZED = "UNINITIALIZED"
        LAUNCH_PREDICT_LOOP = "LAUNCH_PREDICT_LOOP"
        SETUP_SERVICE = "SETUP_SERVICE"
        RUNNING = "RUNNING"
        SHUTTING_DOWN = "SHUTTING_DOWN"
        STOPPED = "STOPPED"

    def __init__(self, service: Service, async_enable: bool = True):
        self.app = FastAPI(lifespan=lambda _: self.prepare_for_serving())
        self.service: Service = service
        self._state = self.State.UNINITIALIZED
        self.is_async = async_enable
        self._request_queue: multiprocessing.Queue[Any] = None
        self._predict_loop_watchdog_task: asyncio.Task = None

        self.response_headers = {
            "OCTOAI_REPLICA_NAME": os.environ.get("OCTOAI_REPLICA_NAME", ""),
        }

        # Build Pydantic models for input/ouput for /infer route
        Input = inspect_input_types(service)
        Output = inspect_output_types(service.infer)

        # Used to read from the request queue in async mode
        async def _pipe_reader(read: Connection):
            """Async multiprocessing.Pipe reader.

            :param read: pipe file handle to read from.
            :return: the contents of the pipe when read.
            """
            data_available = asyncio.Event()
            asyncio.get_event_loop().add_reader(read.fileno(), data_available.set)
            if not read.poll():
                await data_available.wait()
            result = read.recv()
            data_available.clear()
            asyncio.get_event_loop().remove_reader(read.fileno())
            return result

        # Async mode: Put request in the queue for predict loop
        # Sync mode: Call predict method directly
        async def _infer_common(
            service_method: ServiceMethod,
            method_args: Dict[str, Any],
            output_type: Type[BaseModel],
        ) -> Response:
            if not self.is_running:
                return JSONResponse(
                    status_code=HTTPStatus.SERVICE_UNAVAILABLE,
                    content={"status": self.state.name},
                )

            if self.is_async:
                read_conn, write_conn = Pipe()
                start_perf = time.perf_counter_ns()
                request = InferenceRequest(write_conn, service_method, method_args)
                self._request_queue.put(request)
                response = await _pipe_reader(read_conn)
                performance_time_ms = (time.perf_counter_ns() - start_perf) / 1e6
                if isinstance(response, Exception):
                    raise response

                prediction = response.outputs
                inference_time_ms = response.inference_time_ms
            else:
                # track time elapsed in nanoseconds only while app is not asleep
                start_process = time.process_time_ns()
                # track time elapsed in nanoseconds including any sleep time
                start_perf = time.perf_counter_ns()
                infer_fn = getattr(service, service_method.value)
                prediction = infer_fn(**method_args)
                inference_time_ms = (time.process_time_ns() - start_process) / 1e6
                performance_time_ms = (time.perf_counter_ns() - start_perf) / 1e6

            return Response(
                status_code=HTTPStatus.OK,
                headers=self.response_headers,
                media_type="application/json",
                content=output_type(
                    output=prediction,
                    analytics=ResponseAnalytics(
                        inference_time_ms=inference_time_ms,
                        performance_time_ms=performance_time_ms,
                    ),
                ).model_dump_json(),
            )

        # Implementation for form data route.
        # This function signature is dynamically redefined later based on that of
        # Service.infer_form_data() if it is implemented, so that FastAPI
        # can know the parameters and their types when registering the route.
        async def infer_form_data(**kwargs):
            return await _infer_common(
                service_method=ServiceMethod.INFER_FORM_DATA,
                method_args=kwargs,
                output_type=OutputFormData,
            )

        # Add form data route to FastAPI if implemented.
        if implements_form_data(service):
            infer_form_data.__signature__ = transform_form_data_signature(service)
            OutputFormData = inspect_output_types(service.infer_form_data)

            self.app.add_api_route(
                path=service.infer_form_data.__path__,
                endpoint=infer_form_data,
                methods=["POST"],
                response_model=OutputFormData,
            )

        @self.app.get("/healthcheck")
        def health() -> JSONResponse:
            return JSONResponse(
                status_code=HTTPStatus.OK
                if self.is_running
                else HTTPStatus.SERVICE_UNAVAILABLE,
                content={"status": self.state.name, "async_enable": self.is_async},
            )

        @self.app.get("/")
        def root() -> JSONResponse:
            return JSONResponse(
                status_code=HTTPStatus.OK,
                content={
                    "docs": "/docs",
                    "openapi": "/openapi.json",
                },
            )

        @self.app.post(
            "/infer",
            response_model=Output,
        )
        async def infer(request: Input) -> Response:
            return await _infer_common(
                service_method=ServiceMethod.INFER_JSON,
                method_args={k: v for k, v in request},
                output_type=Output,
            )

    @property
    def is_running(self):
        """True when this server instance can serve a request."""
        return self.state == self.State.RUNNING

    @property
    def state(self):
        """Get the status of this server."""
        return self._state

    @state.setter
    def state(self, new_state):
        """Set the status of the server, and log transition."""
        _LOG.info("status: %s -> %s", self._state, new_state)
        self._state = new_state

    @asynccontextmanager
    async def prepare_for_serving(self):
        """Context manager that should surround all serving.

        This is intended to be used as an ASGI application's lifetime handler.
        """
        assert self.state in (
            self.State.UNINITIALIZED,
            self.State.STOPPED,
        ), f"prepare_for_serving: status not UNINITIALIZED or STOPPED: {self.state}"

        _LOG.info("lifecycle: on_server_startup")
        self.service.on_server_startup()

        if self.is_async:
            self.state = self.State.LAUNCH_PREDICT_LOOP
            # load_into_memory is handled in predict loop so that subprocess
            # does all GPU access.
            self._start_predict_loop()
        else:
            self.state = self.State.SETUP_SERVICE
            self.service.setup()
            self.state = self.State.RUNNING

        yield

        self.state = self.State.SHUTTING_DOWN

        if self.is_async:
            self._stop_predict_loop()

        _LOG.info("lifecycle: on_server_shutdown")
        self.service.on_server_shutdown()
        self.state = self.State.STOPPED

    async def _check_predict_loop(self):
        if not self._predict_process.is_alive():
            self._start_predict_loop()

    def _start_predict_loop(self):
        context = multiprocessing.get_context("spawn")
        if not self._request_queue:
            # Only need to create this queue once. This function may be called
            # multiple times if the predict loop dies.
            self._request_queue = context.Queue()

        self._predict_process = context.Process(
            target=_predict_loop,
            name="_predict_loop",
            args=(self.service, self._request_queue),
        )
        self._predict_process.start()

        read_conn, write_conn = Pipe()

        def _remove_reader():
            asyncio.get_event_loop().remove_reader(read_conn)
            read_conn.close()

        def read_startup_message():
            try:
                value = read_conn.recv()
            except EOFError:
                asyncio.get_event_loop().remove_reader(read_conn)
                return

            if isinstance(value, Exception):
                _LOG.error("predict loop died during setup", exc_info=value)
                # Though unconventional, this appears the be the proper way to cleanly
                # terminate uvicorn. It's not easy to get ahold of
                # uvicorn.Server.handle_exit() from here.
                self._stop_predict_loop()
                os.kill(os.getpid(), signal.SIGTERM)

            if value == _LOAD_MODEL_MESSAGE:
                if self.state != self.State.LAUNCH_PREDICT_LOOP:
                    _LOG.error(
                        "_read_predict_loop_startup_message: %s message: "
                        "not in expected state: %s",
                        value,
                        self.state,
                    )
                    _remove_reader()
                    return

                self.state = self.State.SETUP_SERVICE
            elif value == _MODEL_LOADED_MESSAGE:
                if self.state != self.State.SETUP_SERVICE:
                    _LOG.error(
                        "_read_predict_loop_startup_message: %s message: "
                        "not in expected state: %s",
                        value,
                        self.state,
                    )
                    _remove_reader()
                    return

                self.state = self.State.RUNNING
                self._predict_loop_watchdog_task = asyncio.create_task(
                    self._predict_loop_watchdog(  # noqa
                        _PREDICT_LOOP_WATCHDOG_SECONDS, self._check_predict_loop  # noqa
                    )
                )
                _remove_reader()

        asyncio.get_event_loop().add_reader(read_conn.fileno(), read_startup_message)
        self._request_queue.put(write_conn)

    def _stop_predict_loop(self):
        if not self._request_queue:
            return

        if self._predict_loop_watchdog_task is not None:
            self._predict_loop_watchdog_task.cancel()
        self._request_queue.put(_TERMINATE_PREDICT_LOOP_REQUEST)
        self._predict_process.join()
        self._request_queue = None

    async def _predict_loop_watchdog(self, interval, periodic_function):
        while True:
            await asyncio.gather(
                asyncio.sleep(interval),
                periodic_function(),
            )

    def get_api_schema(self) -> Dict[str, Any]:
        """Return the Open API schema for the underlying service."""
        return self.app.openapi()

    def get_usage_examples(
        self, format: documenter.DocumentationFormat
    ) -> MustacheGenerationResult:
        """Return the Mustache generation result for the underlying service."""
        return documenter.generate_usage_examples(self.service, format)

    def store_assets(self) -> None:
        """Run service store assets."""
        self.service.store_assets()

    def run(self, port: int, timeout_keep_alive: int):
        """Run the server exposing the underlying service."""
        uvicorn.run(
            self.app,
            host="0.0.0.0",
            port=port,
            timeout_keep_alive=timeout_keep_alive,
            lifespan="on",
        )


def load_service(module_name: str, class_name: Optional[str] = None) -> Service:
    """Load a class from service implementation."""
    try:
        module = importlib.import_module(module_name)

        if class_name is not None:
            # if service class is provided, instantiate it
            class_ = getattr(module, class_name)
        else:
            # if service class not provided, look for it
            class_ = None
            for name, class_obj in inspect.getmembers(module, inspect.isclass):
                for class_base in class_obj.__bases__:
                    if (
                        class_base.__module__ == _OCTOAI_SERVICE_MODULE
                        and class_base.__name__ == _OCTOAI_BASE_SERVICE_CLASS
                    ):
                        class_ = class_obj
                        break

            if class_ is None:
                raise ValueError(
                    f"Module '{module_name}' contains no classes extending "
                    f"base '{_OCTOAI_SERVICE_MODULE}.{_OCTOAI_BASE_SERVICE_CLASS}'"
                )

        _LOG.info(f"Using service in {module_name}.{class_.__name__}.")

        return class_()
    except ModuleNotFoundError:
        error_msg = f"Module '{module_name}' not found. "
        if module_name == "service":
            error_msg += "Ensure your service is defined in service.py."
        raise ValueError(error_msg)


def _load_server(ctx, async_enable: bool = True) -> Server:
    service = load_service(
        ctx.parent.params["service_module"],
        class_name=ctx.parent.params["service_class"],
    )
    return Server(service, async_enable=async_enable)


@click.group(name="server")
@click.option(
    "--log-level",
    type=click.Choice(["ERROR", "WARNING", "INFO", "DEBUG"], case_sensitive=False),
    default="INFO",
    envvar="OCTOAI_LOG_LEVEL",
)
@click.option("--service-module", default="service")
@click.option("--service-class", default=None)
@click.pass_context
def server(ctx, log_level, service_module, service_class):
    """CLI for OctoAI server."""
    logging.basicConfig(
        level=log_level,
        stream=sys.stderr,
        format="%(asctime)s %(levelname)s %(filename)s:%(lineno)d %(message)s",
    )
    click.echo("octoai server")
    ctx.ensure_object(dict)


@server.command()
@click.option("--output-file", default=None)
@click.pass_context
def api_schema(ctx, output_file):
    """Generate OpenAPI schema for the given service."""
    _LOG.info("api-schema")
    server: Server = _load_server(ctx)
    schema = server.get_api_schema()

    if output_file:
        with open(output_file, "w") as f:
            json.dump(schema, f, indent=2)
    else:
        click.echo(json.dumps(schema, indent=2))


@server.command(hidden=True)
@click.option("--no-render", is_flag=True, default=False)
@click.option(
    "--format",
    required=True,
    type=click.Choice(["curl", "python"], case_sensitive=False),
)
@click.option("--output-file", default=None)
@click.pass_context
def generate_usage_examples(ctx, no_render, format, output_file):
    """Generate client usage examples for the given service."""
    _LOG.info("generate-usage-examples")

    def convert_format(f: str) -> documenter.DocumentationFormat:
        if f.lower() == "curl":
            return documenter.DocumentationFormat.CURL
        elif f.lower() == "python":
            return documenter.DocumentationFormat.PYTHON
        else:
            raise ValueError(f"format '{f}' is not supported")

    server: Server = _load_server(ctx)
    generation_result = server.get_usage_examples(convert_format(format))

    if not no_render:
        rendered_result = chevron.render(
            generation_result.generated_template, generation_result.template_hash
        )

        if output_file:
            with open(output_file, "w") as f:
                f.write(rendered_result)
        else:
            click.echo(rendered_result)
    else:
        if output_file:
            with open(output_file, "w") as f:
                json.dump(dataclasses.asdict(generation_result), f, indent=2)
        else:
            click.echo(json.dumps(dataclasses.asdict(generation_result), indent=2))


@server.command()
@click.pass_context
def setup_service(ctx):
    """Run the setup code for the given service."""
    _LOG.info("setup-service")
    server: Server = _load_server(ctx, async_enable=False)
    server.service.setup()


@server.command()
@click.pass_context
@click.option("--port", type=int, default=8080)
@click.option("--async-enable", default=True)
@click.option(
    "--timeout-keep-alive",
    type=int,
    default=900,
    help="Connection keep alive timeout in seconds",
)
def run(ctx, port, async_enable, timeout_keep_alive):
    """Run the server for the given service."""
    _LOG.info("run")
    maybe_set_volume_environment_variables(ctx)
    server: Server = _load_server(ctx, async_enable=async_enable)
    server.run(port, timeout_keep_alive)


@server.command()
@click.pass_context
@click.option("--check-is-needed", is_flag=True)
def store_assets(ctx, check_is_needed):
    """Run the store_assets code for the given model."""
    _LOG.info("store_assets")
    if check_is_needed:
        sys.exit(
            _STORE_ASSET_NEEDED_CODE
            if is_store_assets_needed(ctx)
            else _STORE_ASSET_NOT_NEEDED_CODE
        )
    maybe_set_volume_environment_variables(ctx)
    server: Server = _load_server(ctx, async_enable=False)
    server.store_assets()


if __name__ == "__main__":
    server(obj={})



"""An AST for the parsed ``octoai.service.Service`` subclass."""

import enum
import functools
import inspect
import sys
import typing
from inspect import Signature
from typing import Any, Callable, Dict, List, Set, Tuple, Type, Union

import typing_extensions
from pydantic import BaseModel, HttpUrl
from pydantic.fields import FieldInfo as PydanticFieldInfo
from pydantic_core import Url

from octoai.service import Service, inspect_input_parameters
from octoai.spec.ast import (
    BaseType,
    CustomObjectType,
    DictionaryType,
    EnumType,
    Field,
    FieldInfo,
    InferenceEndpoint,
    InferenceService,
    ListType,
    MediaKind,
    MediaType,
    NamedField,
    OptionalType,
    PrimitiveType,
    PydanticHttpUrlType,
    SetType,
    SimpleFieldInfo,
    TupleType,
    UnionType,
    UnknownType,
)
from octoai.types import Audio, Image, Video


def parse_service(service_class: Type[Service]) -> InferenceService:
    """Parse the ``Service`` subclass and return the AST of the parsed service."""
    # Make sure this is a subclass of `Service` but not `Service` itself.
    if not issubclass(service_class, Service) or service_class == Service:
        raise ValueError(
            "The provided service module must be a subclass of octoai.Service"
        )

    inference_endpoint = _parse_inference_endpoint(service_class.infer)
    return InferenceService(inference_endpoint=inference_endpoint)


def _parse_inference_endpoint(infer: Callable[..., Any]) -> InferenceEndpoint:
    """Introspect the input/output from the ``infer()`` signature."""
    signature = inspect.signature(infer)
    service_input_parameters = signature.parameters
    # The conditional statement here uses < 2 as the condition because the `Service`
    # subclass must have at least one parameter. Because we are looking at the
    # class-level type, `self` is not bound and thus there is an extra parameter.
    if len(service_input_parameters) < 2:
        raise ValueError("infer() requires type annotations for args")

    # Remove `self`, but now we need to get the parameters again.
    infer = functools.partial(infer, None)
    parameters = inspect.signature(infer).parameters
    parsed_input = _parse_field(inspect_input_parameters(parameters))

    service_output = signature.return_annotation
    if service_output == Signature.empty:
        raise ValueError("infer() requires return type annotation")
    parsed_output = _parse_field(service_output)

    return InferenceEndpoint(inputs=parsed_input, output=parsed_output)


def _parse_field(return_annotation: Any) -> Field:
    """Parse a field in an object."""
    return _parse_type_annotation(return_annotation)


def _parse_type_annotation(annotation: Any) -> Field:
    parameter_type_annotation = type(annotation)

    if typing.get_origin(parameter_type_annotation) is typing_extensions.Annotated or (
        sys.version_info >= (3, 9)
        and typing.get_origin(parameter_type_annotation) is typing.Annotated
    ):
        return _parse_annotated_type_annotation(annotation)
    else:
        return _parse_unit_type_annotation(annotation)


def _parse_unit_type_annotation(annotation: Any) -> Field:
    return Field(
        type=parse_type(annotation),
        info=SimpleFieldInfo(
            default=None,
            description=None,
        ),
    )


# No annotation for the Annotation type because Python 3.8 does not have ``Annotated``.
def _parse_annotated_type_annotation(annotation) -> Field:  # type: ignore[valid-type] # noqa
    # FastAPI requires the first annotation argument to be the actual type. We follow
    # the same requirement here.
    args = typing.get_args(annotation)

    if len(args) < 1:
        raise ValueError(
            "annotated types on the signature in infer() must have the base type"
        )

    base_type = args[0]

    # Search the remaining annotations for Pydantic Field.
    info: FieldInfo = SimpleFieldInfo(default=None, description=None)
    for arg in args[1:]:
        if type(arg) is PydanticFieldInfo:
            info = arg

    return Field(
        type=parse_type(base_type),
        info=info,
    )


def parse_type(data_type: Type[Any]) -> BaseType:
    """Parse the annotated type and return a subclass of ``BaseType`` in the AST."""
    if (
        data_type == str
        or data_type == bytes
        or data_type == int
        or data_type == float
        or data_type == bool
    ):
        return PrimitiveType(type=data_type)
    elif data_type == Video:
        return MediaType(kind=MediaKind.VIDEO)
    elif data_type == Audio:
        return MediaType(kind=MediaKind.AUDIO)
    elif data_type == Image:
        return MediaType(kind=MediaKind.IMAGE)
    elif data_type == HttpUrl or data_type == Url:
        return PydanticHttpUrlType()
    elif inspect.isclass(data_type) and issubclass(data_type, enum.Enum):
        return _parse_enum_type(data_type)
    elif inspect.isclass(data_type) and issubclass(data_type, BaseModel):
        return _parse_custom_object_type(data_type)
    elif type(data_type) is typing._GenericAlias and (  # type: ignore[attr-defined] # noqa
        typing.get_origin(data_type) == list
        or typing.get_origin(data_type) == tuple
        or typing.get_origin(data_type) == dict
        or typing.get_origin(data_type) == set
    ):
        return _parse_aggregation_type(data_type)
    elif (
        isinstance(data_type, typing._GenericAlias)  # type: ignore[attr-defined] # noqa
        and typing.get_origin(data_type) == Union
    ):
        return _parse_union_type(data_type)
    else:
        raise NotImplementedError(
            f"currently the `{data_type.__name__}` type is unsupported"
        )


def _parse_enum_type(enum_type: Type[enum.Enum]):
    variants_and_representations = [
        (member, variant.value) for (member, variant) in enum_type.__members__.items()
    ]
    return EnumType(name=enum_type.__name__, variants=variants_and_representations)


def _parse_aggregation_type(
    aggregation_type: Union[Type[Dict], Type[List], Type[Tuple], Type[Set]]
) -> BaseType:
    args = typing.get_args(aggregation_type)

    if aggregation_type.__origin__ == list:  # type: ignore[union-attr] # noqa
        return ListType(
            member_type=parse_type(args[0]) if len(args) == 1 else UnknownType()
        )
    elif aggregation_type.__origin__ == set:  # type: ignore[union-attr] # noqa
        return SetType(
            member_type=parse_type(args[0]) if len(args) == 1 else UnknownType()
        )
    elif aggregation_type.__origin__ == tuple:  # type: ignore[union-attr] # noqa
        members = [parse_type(arg) for arg in args]
        return TupleType(members=members if members else [UnknownType()])
    elif aggregation_type.__origin__ == dict:  # type: ignore[union-attr] # noqa
        key_type: BaseType = UnknownType()
        value_type: BaseType = UnknownType()

        if len(args) > 0:
            key_type = parse_type(args[0])
            value_type = parse_type(
                args[1]
            )  # Python already checks if ValueType is specified.

        return DictionaryType(key_type=key_type, value_type=value_type)
    else:
        raise ValueError(
            f"the `{aggregation_type.__name__}` is not an aggregation type"
        )


def _parse_union_type(union_type: Type[Union[Any]]) -> BaseType:
    args = typing.get_args(union_type)
    if len(args) == 2 and args[1] is type(None):
        return OptionalType(some_type=parse_type(args[0]))
    return UnionType(members=[parse_type(arg) for arg in args])


def _parse_pydantic_field_info(name: str, field_info: PydanticFieldInfo) -> Field:
    # Pydantic’s ``FieldInfo`` provides a trove of information. Just attach it.
    return NamedField(
        name=name,
        type=parse_type(field_info.annotation),
        info=field_info,
    )


def _parse_custom_object_type(custom_object_type: Type[BaseModel]) -> BaseType:
    model_fields = custom_object_type.model_fields

    fields = {}
    for field_name, field in model_fields.items():
        parsed_field = _parse_pydantic_field_info(field_name, field)
        fields[field_name] = parsed_field

    return CustomObjectType(name=custom_object_type.__name__, fields=fields)



# coding: utf-8

"""
    FastAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Dict, Optional, Union

from pydantic.v1 import BaseModel, Field, StrictBool, StrictFloat, StrictInt, StrictStr


class LegacyImageGenerationRequest(BaseModel):
    """
    Model input data type.  # noqa: E501
    """

    prompt: StrictStr = Field(...)
    prompt_2: Optional[StrictStr] = None
    negative_prompt: Optional[StrictStr] = None
    negative_prompt_2: Optional[StrictStr] = None
    model: Optional[StrictStr] = None
    loras: Optional[Dict[str, Union[StrictFloat, StrictInt]]] = None
    sampler: Optional[StrictStr] = "DDIM"
    height: Optional[StrictInt] = 1024
    width: Optional[StrictInt] = 1024
    cfg_scale: Optional[Union[StrictFloat, StrictInt]] = 12.0
    steps: Optional[StrictInt] = 30
    num_images: Optional[StrictInt] = 1
    seed: Optional[StrictInt] = None
    init_image: Optional[StrictStr] = None
    strength: Optional[Union[StrictFloat, StrictInt]] = 0.8
    style_preset: Optional[StrictStr] = None
    use_refiner: Optional[StrictBool] = True
    high_noise_frac: Optional[Union[StrictFloat, StrictInt]] = 0.8
    enable_safety: Optional[StrictBool] = False
    output_type: Optional[StrictStr] = "jpg"
    __properties = [
        "prompt",
        "prompt_2",
        "negative_prompt",
        "negative_prompt_2",
        "model",
        "loras",
        "sampler",
        "height",
        "width",
        "cfg_scale",
        "steps",
        "num_images",
        "seed",
        "init_image",
        "strength",
        "style_preset",
        "use_refiner",
        "high_noise_frac",
        "enable_safety",
        "output_type",
    ]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> LegacyImageGenerationRequest:
        """Create an instance of LegacyImageGenerationRequest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> LegacyImageGenerationRequest:
        """Create an instance of LegacyImageGenerationRequest from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return LegacyImageGenerationRequest.parse_obj(obj)

        _obj = LegacyImageGenerationRequest.parse_obj(
            {
                "prompt": obj.get("prompt"),
                "prompt_2": obj.get("prompt_2"),
                "negative_prompt": obj.get("negative_prompt"),
                "negative_prompt_2": obj.get("negative_prompt_2"),
                "model": obj.get("model"),
                "loras": obj.get("loras"),
                "sampler": obj.get("sampler")
                if obj.get("sampler") is not None
                else "DDIM",
                "height": obj.get("height") if obj.get("height") is not None else 1024,
                "width": obj.get("width") if obj.get("width") is not None else 1024,
                "cfg_scale": obj.get("cfg_scale")
                if obj.get("cfg_scale") is not None
                else 12.0,
                "steps": obj.get("steps") if obj.get("steps") is not None else 30,
                "num_images": obj.get("num_images")
                if obj.get("num_images") is not None
                else 1,
                "seed": obj.get("seed"),
                "init_image": obj.get("init_image"),
                "strength": obj.get("strength")
                if obj.get("strength") is not None
                else 0.8,
                "style_preset": obj.get("style_preset"),
                "use_refiner": obj.get("use_refiner")
                if obj.get("use_refiner") is not None
                else True,
                "high_noise_frac": obj.get("high_noise_frac")
                if obj.get("high_noise_frac") is not None
                else 0.8,
                "enable_safety": obj.get("enable_safety")
                if obj.get("enable_safety") is not None
                else False,
                "output_type": obj.get("output_type")
                if obj.get("output_type") is not None
                else "jpg",
            }
        )
        return _obj



from __future__ import annotations

# coding: utf-8

"""
    asset-lake-api-server

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import json
import pprint
import re  # noqa: F401
from enum import Enum


class FileExtension(str, Enum):
    """
    FileExtension
    """

    """
    allowed enum values
    """
    UNKNOWN = "unknown"
    PNG = "png"
    JPG = "jpg"
    JPEG = "jpeg"
    TXT = "txt"

    @classmethod
    def from_json(cls, json_str: str) -> FileExtension:
        """Create an instance of FileExtension from a JSON string"""
        return FileExtension(json.loads(json_str))



"""
Type definitions to help communicate with endpoints.

These type definitions help with routine format conversions
that are necessary when transferring binary files (such as
images or audio) over HTTP. These type definitions can be useful
both when creating endpoints (implementing ``octoai.service.Service``
as directed by the ``octoai`` command-line interface) and when
communicating with live endpoints using the client SDK.
"""

import base64
import importlib
import os
import sys
import tempfile
from io import BytesIO
from types import ModuleType
from typing import Any, Dict, Iterator, List, Tuple, Union, cast

import httpx
from fastapi import UploadFile
from numpy.typing import ArrayLike
from PIL import Image as PImage
from pydantic import BaseModel, Field, HttpUrl, field_serializer


def _import_soundfile() -> ModuleType:
    """Attempt to import the soundfile module."""
    if "soundfile" in sys.modules:
        return sys.modules["soundfile"]

    try:
        return importlib.import_module("soundfile")
    except OSError:
        raise Exception(
            "Can't import the 'soundfile' module. "
            "For Linux, try: sudo apt install libsndfile1. "
            "For Mac, try: brew install libsndfile."
        )


def _import_imageio() -> ModuleType:
    """Attempt to import the imageio.v3 module."""
    if "imageio.v3" in sys.modules:
        return sys.modules["imageio.v3"]

    try:
        return importlib.import_module("imageio.v3")
    except OSError:
        raise Exception(
            "Can't import the 'imageio.v3' module. "
            "Try reinstalling octoai-sdk with the following command: "
            "pip install octoai-sdk[video]"
        )


class MediaUrl(BaseModel):
    """URL reference to a media file."""

    url: HttpUrl

    @field_serializer("url")
    def url2str(self, val) -> str:
        """Serialize HttpUrl to str."""
        return str(val)


class Media(BaseModel):
    """Base class for media files.

    The `Media` class is a wrapper for either binary media content
    or a URL reference to a media file. When an instance contains
    binary media data, it is stored and transferred over HTTP as a
    base64-encoded string. When an instance contains a URL reference,
    the URL is stored and transferred over HTTP as a string. The
    URL representation can be more advantageous when the media file is
    already hosted on a server, as it avoids the overhead of base64
    encoding and decoding.

    The `Media` class is the base class for the `Image`, `Audio`, and
    `Video` classes, which contain additional convenience methods to
    read and write media files in common formats. These subclasses
    should be used instead of the `Media` in most cases.
    """

    data: Union[bytes, MediaUrl] = Field(
        description="URL reference or base64-encoded data"
    )

    def __init__(self, data: Union[bytes, MediaUrl]):
        super().__init__(data=data)

    @classmethod
    def from_base64(cls, b64: bytes) -> "Media":
        """Create from base64-encoded data, such as that returned from an HTTP call.

        See also ``Media.from_endpoint_response()``.

        :param b64: base64-encoded data representing a media file
        "type b64: bytes
        :return: ``Media`` instance
        :rtype: :class:`Media`
        """
        return cls(data=b64)

    @classmethod
    def from_endpoint_response(cls, resp_dict: Dict[str, Any], key: str) -> "Media":
        """Create from an endpoint response, such as an endpoint that produces media.

        :param resp_dict: a response from an OctoAI endpoint that produces media
        :type resp_dict: Dict[str, Any]
        :param key: the key name in the response that contains the media
        :return: `Media` object
        :rtype: :class:`Media`
        """
        if key in resp_dict:
            return cls(**resp_dict[key])
        elif "output" in resp_dict and key in resp_dict["output"]:
            return cls(**resp_dict["output"][key])

        raise ValueError(f"Key {key} not found in response")

    @classmethod
    def from_file(cls, file_name: str) -> "Media":
        """Create media from local file.

        :param file_name: path to local media file
        :type file_name: str
        :raises ValueError: file_name not found at provided path
        :return: `Media` object
        :rtype: :class:`Media`
        """
        if not os.path.isfile(file_name):
            raise ValueError(f"File {file_name} does not exist")

        with open(file_name, "rb") as fd:
            return cls(data=base64.b64encode(fd.read()))

    @classmethod
    def from_url(cls, url: str, b64=False, follow_redirects=False) -> "Media":
        """Create media from URL.

        :param url: URL to media file
        :type url: str
        :param b64: whether to base64-encode the media file, defaults to False
        :type b64: bool, optional
        :param follow_redirects: whether to follow redirects, defaults to False
        :type follow_redirects: bool, optional
        :rtype: :class:`Media`
        """
        if b64:
            resp = httpx.get(url, follow_redirects=follow_redirects)
            if resp.status_code != 200:
                raise ValueError(f"status {resp.status_code} ({url})")

            return cls(data=base64.b64encode(resp.content))
        else:
            return cls(data=MediaUrl(url=HttpUrl(url)))

    def download(self, follow_redirects, file_name):
        """Download media URL to local file.

        :param follow_redirects: whether to follow redirects
        :type follow_redirects: bool
        :param file_name: path to local file
        :type file_name: str
        """
        if isinstance(self.data, bytes):
            return
        elif isinstance(self.data, MediaUrl):
            with httpx.stream(
                "GET",
                url=str(self.data.url),
                follow_redirects=follow_redirects,
            ) as resp:
                if resp.status_code != 200:
                    raise ValueError(
                        f"status {resp.status_code} ({str(self.data.url)})"
                    )
                for chunk in resp.iter_bytes():
                    with open(file_name, "ab") as f:
                        f.write(chunk)

    def default_download(self, file_name: str):
        """Download media URL to a file using the default settings.

        Override this to change the download behavior.

        :param file_name: path to local file
        :type file_name: str
        """
        self.download(follow_redirects=False, file_name=file_name)

    def to_file(self, file_name: str):
        """Write media to local file.

        :param file_name: path to local file
        :type file_name: str
        """
        if isinstance(self.data, bytes):
            with open(file_name, "wb") as fd:
                fd.write(base64.b64decode(self.data))
        elif isinstance(self.data, MediaUrl):
            self.default_download(file_name=file_name)

    def to_bytes(self) -> bytes:
        """Convert media to bytes.

        :return: media as bytes
        :rtype: bytes
        """
        if isinstance(self.data, bytes):
            return base64.b64decode(self.data)

        if isinstance(self.data, MediaUrl):
            with tempfile.NamedTemporaryFile() as f:
                self.default_download(file_name=f.name)
                with open(f.name, "rb") as fd:
                    return fd.read()


class Image(Media):
    """Image helpers for models that accept or return images.

    The `Image` class is a wrapper for either binary image content
    or a URL reference to an image file. When an instance contains
    binary image data, it is stored and transferred over HTTP as a
    base64-encoded string. When an instance contains a URL reference,
    the URL is stored and transferred over HTTP as a string. The
    URL representation can be more advantageous when the image file is
    already hosted on a server, as it avoids the overhead of base64
    encoding and decoding.

    The `Image` class contains additional convenience methods to
    read and write image files in common formats.
    """

    @classmethod
    def from_pil(cls, image_pil: PImage, format="JPEG") -> "Image":
        """Create from Pillow image object.

        A file format is required since the Pillow image object is
        serialized to a binary image file. The default is "JPEG".

        :param image_pil: image in PIL format
        :type image_pil: PIL.Image
        :param format: target file format, defaults to "JPEG"
        :type format: str, optional
        :return: `Image` object
        :rtype: :class:`Image`
        """
        buffer = BytesIO()
        image_pil.save(buffer, format=format)
        return cls(data=base64.b64encode(buffer.getvalue()))

    def to_pil(self) -> PImage:
        """Convert to PIL Image.

        :return: Pillow image object
        :rtype: PIL.Image
        """
        return PImage.open(BytesIO(self.to_bytes()))

    def is_valid(self):
        """Check if this is a valid image.

        :return: True if valid, False if invalid
        :rtype: bool
        """
        try:
            self.to_pil().verify()
            return True
        except Exception:
            return False


class Audio(Media):
    """Audio helpers for models that accept or return audio.

    The `Audio` class is a wrapper for either binary audio content
    or a URL reference to an audio file. When an instance contains
    binary audio data, it is stored and transferred over HTTP as a
    base64-encoded string. When an instance contains a URL reference,
    the URL is stored and transferred over HTTP as a string. The
    URL representation can be more advantageous when the audio file is
    already hosted on a server, as it avoids the overhead of base64
    encoding and decoding.

    The `Audio` class contains additional convenience methods to
    read and write audio files in common formats.
    """

    @classmethod
    def from_numpy(cls, data: ArrayLike, sample_rate: int, format="WAV") -> "Audio":
        """Create from a numpy array.

        The first dimension of the array should represent audio frames (samples),
        while the second dimension should represent audio channels.

        A file format and a sample rate are needed since the audio data is
        serialized to a binary audio file. The default format is "WAV", and the
        sample rate is required.

        :param data: numpy array with audio data (frames x channels)
        :type data: ArrayLike
        :param sample_rate: samples per second taken to create signal
        :type sample_rate: int
        :param format: target format, defaults to "WAV"
        :type format: str, optional
        :return: Audio object
        :rtype: :class:`Audio`
        """
        soundfile = _import_soundfile()
        buffer = BytesIO()
        soundfile.write(buffer, data=data, samplerate=sample_rate, format=format)
        return cls(data=base64.b64encode(buffer.getvalue()))

    def to_numpy(self) -> Tuple[ArrayLike, int]:
        """Convert to numpy array.

        :return: numpy array representation (frames x channels)
        :rtype: Tuple[ArrayLike, int]
        """
        soundfile = _import_soundfile()
        fd = BytesIO(self.to_bytes())
        data, sample_rate = soundfile.read(fd)
        return (data, sample_rate)

    def is_valid(self):
        """Check if this is a valid audio.

        :return: True if it's valid, false if not.
        :rtype: bool
        """
        try:
            self.to_numpy()
            return True
        except Exception:
            return False


class Video(Media):
    """Video helpers for models that accept or return video.

    The `Video` class is a wrapper for either binary video content
    or a URL reference to a video file. When an instance contains
    binary video data, it is stored and transferred over HTTP as a
    base64-encoded string. When an instance contains a URL reference,
    the URL is stored and transferred over HTTP as a string. The
    URL representation can be more advantageous when the video file is
    already hosted on a server, as it avoids the overhead of base64
    encoding and decoding.

    The `Video` class contains additional convenience methods to
    read and write video files in common formats.
    """

    @classmethod
    def from_numpy(
        cls,
        video_frames: List[ArrayLike],
        codec="libx264",
        pixel_format="yuv420p",
        fps=30,
    ) -> "Video":
        """Create from NumPy video frames.

        :param video_frames: list of NumPy arrays representing video frames
        :type video_frames: List[ArrayLike]
        :param codec: FFMPEG video codec, defaults to 'libx264'
        :type codec: str, optional
        :param pixel_format: FFMPEG pixel format, defaults to 'yuv420p'
        :type pixel_format: str, optional
        :param fps: frames per second, defaults to 30
        :type fps: int, optional
        :return: Video object
        :rtype: :class:`Video`
        """
        iiov3 = _import_imageio()

        with tempfile.NamedTemporaryFile(suffix=".mp4") as f:
            with iiov3.imopen(f.name, "w", plugin="pyav") as vf:
                vf.init_video_stream(codec=codec, fps=fps, pixel_format=pixel_format)

                for frame in video_frames:
                    vf.write_frame(frame)  # type: ignore[arg-type]

            return cast(Video, cls.from_file(f.name))

    def to_numpy_iterator(self) -> Tuple[Iterator[ArrayLike], str]:
        """Convert to a NumPy iterator.

        :return: tuple of iterator of NumPy arrays representing video frames
            and the file name of the temporary file. Users of this method
            should delete the temporary file when the iterator is no longer
            needed to avoid filling up the disk.
        :rtype: Tuple[Iterator[ArrayLike], str]
        """
        iiov3 = _import_imageio()

        # return also file name so user can delete it when they're done
        with tempfile.NamedTemporaryFile(suffix=".mp4", delete=False) as f:
            self.to_file(f.name)
            return iiov3.imiter(f.name), f.name

    def to_numpy(self) -> ArrayLike:
        """Convert to a NumPy array.

        :return: NumPy array representing video frames
        :rtype: ArrayLike
        """
        iiov3 = _import_imageio()

        with tempfile.NamedTemporaryFile(suffix=".mp4") as f:
            self.to_file(f.name)
            return iiov3.imread(f.name)

    def is_valid(self) -> bool:
        """Check if this is a valid video.

        :return: True if valid, False if invalid
        :rtype: bool
        """
        iiov3 = _import_imageio()
        from imageio.core.request import InitializationError

        with tempfile.NamedTemporaryFile(suffix=".mp4") as f:
            self.to_file(f.name)
            fname = f.name

            try:
                metadata = iiov3.immeta(fname, plugin="pyav")
            except OSError:
                # audio files cause this error
                return False
            except InitializationError:
                # incompatible files cause this error
                return False

            return "video_format" in metadata


# create namespace for future extensions
class File(UploadFile):
    """File class for file uploads in form-data endpoints."""

    pass



"""Base classes for generation of template."""

from abc import ABC, abstractmethod
from copy import copy
from dataclasses import dataclass
from typing import Dict, List, Optional, Union, cast

from pydantic.fields import FieldInfo as PydanticFieldInfo
from pydantic_core import PydanticUndefined

from octoai.spec.ast import (
    BaseType,
    CustomObjectType,
    DictionaryType,
    EnumType,
    Field,
    FieldInfo,
    InferenceService,
    ListType,
    MediaType,
    OptionalType,
    PrimitiveType,
    PydanticHttpUrlType,
    SetType,
    SimpleFieldInfo,
    SpecialType,
    TupleType,
    UnionType,
    UnknownType,
)


class Buffer:
    """Holds a string of the generated template and manages the indentation level."""

    _INDENTATION_LEVEL = 4

    def __init__(self, indentation_level: int = 0):
        self._generated_strings = ""
        self._indentation_level = indentation_level

    def increase_indent(self):
        """Increase the indent by one level."""
        self._indentation_level += 1

    def decrease_indent(self):
        """Decrease the indent by one level."""
        self._indentation_level -= 1

    def print_line(self, line: str, indented: bool = True):
        """Append the line to the end of the buffer.

        :param line: the line to be added
        :param indented: whether to add indentation to the start of the line
        """
        self._generated_strings += self._get_indentation_string() + line
        self.new_line(indented=indented)

    def print_string(self, s: str):
        """Append the string to the end of the buffer.

        :param s: the string to be added
        """
        self._generated_strings += s

    def new_line(self, indented: bool = True):
        """Add a new line.

        :param indented: whether to add indentation to the start of the line
        """
        self._generated_strings += "\n"
        if indented:
            self._generated_strings += self._get_indentation_string()

    def _get_indentation_string(self) -> str:
        return " " * (Buffer._INDENTATION_LEVEL * self._indentation_level)

    @property
    def string(self) -> str:
        """Return the contents."""
        return str(self)

    def __str__(self):
        """Return the contents."""
        return self._generated_strings


TemplateRegistryHashType = Union[str, Dict[str, "TemplateRegistryHashType"]]


class MustacheTemplateRegistry:
    """Store for all the Mustache tags and hashes."""

    def __init__(
        self,
        base_hash: Dict[str, TemplateRegistryHashType] = None,
        field_prefix: str = "",
    ):
        self._hash = base_hash if base_hash else {}
        self._field_prefix = field_prefix

    def register_example(self, key_path: List[str], value: str):
        """Register a new value at the key path.

        :param key_path: a key path referencing the object.
        :param value: the value for the object in the key path.
        """
        path = copy(key_path)

        # Iteratively update or assign the data.
        current_dict = self._hash
        for idx, key in enumerate(path):
            if idx == len(key_path) - 1:
                current_dict[key] = value
                return

            if key not in current_dict:
                current_dict[key] = {}
            current_dict = cast(Dict[str, TemplateRegistryHashType], current_dict[key])

    @property
    def hash(self) -> Dict[str, TemplateRegistryHashType]:
        """Return the hash."""
        return self._hash

    def _get_field_name(self, key_path: List[str]) -> str:
        return self._field_prefix + ".".join(key_path)

    def get_field_template_tag(self, key_path: List[str]) -> str:
        """Get the template tag for the object at key path.

        Triple braces are used to disable Mustache’s escaping for HTML symbols like
        '<' and '>'.

        :param key_path: the referenced key path.
        """
        return "{{{" + self._get_field_name(key_path) + "}}}"

    def get_field_opening_format_tag(self, key_path: List[str]) -> str:
        """Get the opening tag for the object at key path.

        :param key_path: the referenced key path.
        """
        return "{{#" + self._get_field_name(key_path) + "}}"

    def get_field_closing_format_tag(self, key_path: List[str]) -> str:
        """Get the closing tag for the object at key path.

        :param key_path: the referenced key path.
        """
        return "{{/" + self._get_field_name(key_path) + "}}"


@dataclass
class MustacheGenerationResult:
    """Generated Mustache template string and the hashes."""

    generated_template: str
    template_hash: Dict[str, TemplateRegistryHashType]


class ASTVisitor(ABC):
    """Visitor for visiting all the AST nodes."""

    def visit_base(
        self,
        base_type: BaseType,
        field_info: FieldInfo,
    ):
        """Visit the base type.

        :param base_type: the base type.
        :param field_info: information about the field.
        """
        if isinstance(base_type, SpecialType):
            self.visit_special(cast(SpecialType, base_type), field_info)
        elif type(base_type) is CustomObjectType:
            self.visit_custom_object(cast(CustomObjectType, base_type), field_info)
        elif type(base_type) is PrimitiveType:
            self.visit_primitive(cast(PrimitiveType, base_type), field_info)
        elif type(base_type) is UnionType:
            self.visit_union(cast(UnionType, base_type), field_info)
        elif type(base_type) is ListType:
            self.visit_list(cast(ListType, base_type), field_info)
        elif type(base_type) is TupleType:
            self.visit_tuple(cast(TupleType, base_type), field_info)
        elif type(base_type) is SetType:
            self.visit_set(cast(SetType, base_type), field_info)
        elif type(base_type) is DictionaryType:
            self.visit_dict(cast(DictionaryType, base_type), field_info)
        elif isinstance(base_type, EnumType):
            self.visit_enum(cast(EnumType, base_type), field_info)
        elif type(base_type) is UnknownType:
            self.visit_unknown(cast(UnknownType, base_type), field_info)
        elif type(base_type) is OptionalType:
            self.visit_optional(cast(OptionalType, base_type), field_info)
        else:
            raise NotImplementedError(
                f"the type '{base_type}' is currently unsupported"
            )

    def visit_special(
        self,
        special_type: SpecialType,
        field_info: FieldInfo,
    ):
        """Visit the special type.

        :param special_type: the special type.
        :param field_info: information about the field.
        """
        if type(special_type) is MediaType:
            self.visit_media(cast(MediaType, special_type), field_info)
        elif type(special_type) is PydanticHttpUrlType:
            self.visit_pydantic_http_url(
                cast(PydanticHttpUrlType, special_type), field_info
            )
        else:
            raise NotImplementedError(
                f"the type '{type(special_type).__name__}' is currently unsupported"
            )

    def visit_primitive(
        self,
        primitive_type: PrimitiveType,
        field_info: FieldInfo,
    ):
        """Visit the primitive type.

        :param primitive_type: the primitive type.
        :param field_info: information about the field.
        """

    def visit_union(self, union_type: UnionType, field_info: FieldInfo):
        """Visit the union type.

        :param union_type: the union type.
        :param field_info: information about the field.
        """

    def visit_list(
        self,
        list_type: ListType,
        field_info: FieldInfo,
    ):
        """Visit the list type.

        :param list_type: the list type.
        :param field_info: information about the field.
        """
        return

    def visit_tuple(
        self,
        tuple_type: TupleType,
        field_info: FieldInfo,
    ):
        """Visit the tuple type.

        :param tuple_type: the tuple type.
        :param field_info: information about the field.
        """

    def visit_set(
        self,
        set_type: SetType,
        field_info: FieldInfo,
    ):
        """Visit the set type.

        :param set_type: the set type.
        :param field_info: information about the field.
        """

    def visit_dict(
        self,
        dict_type: DictionaryType,
        field_info: FieldInfo,
    ):
        """Visit the dict type.

        :param dict_type: the dict type.
        :param field_info: information about the field.
        """

    def visit_enum(
        self,
        enum_type: EnumType,
        field_info: FieldInfo,
    ):
        """Visit the enum type.

        :param enum_type: the enum type.
        :param field_info: information about the field.
        """

    def visit_unknown(self, unknown_type: UnknownType, field_info: FieldInfo):
        """Visit the unknown type.

        :param unknown_type: the unknown type.
        :param field_info: information about the field.
        """

    def visit_optional(self, optional_type: OptionalType, field_info: FieldInfo):
        """Visit the optional type.

        :param optional_type: the optional type.
        :param field_info: information about the field.
        """

    def visit_media(
        self,
        media_type: MediaType,
        field_info: FieldInfo,
    ):
        """Visit the media type.

        :param media_type: the media type.
        :param field_info: information about the field.
        """

    def visit_pydantic_http_url(
        self,
        pydantic_http_url_type: PydanticHttpUrlType,
        field_info: FieldInfo,
    ):
        """Visit the Pydantic HttpUrl type.

        :param pydantic_http_url_type: the Pydantic HttpUrl type.
        :param field_info: information about the field.
        """

    def visit_custom_object(
        self,
        custom_object_type: CustomObjectType,
        field_info: FieldInfo,
    ):
        """Visit the custom object type.

        :param custom_object_type: The custom object type.
        :param field_info: information about the field.
        """


class MustacheTemplateGeneratingVisitor(ASTVisitor, ABC):
    """A subclass of ``ASTVisitor`` that stores states for Mustache templates."""

    def __init__(
        self,
        base_buffer: Optional[Buffer] = None,
        template_registry: Optional[MustacheTemplateRegistry] = None,
    ):
        super().__init__()
        self._buffer = base_buffer if base_buffer is not None else Buffer()
        self._template_registry = (
            template_registry
            if template_registry is not None
            else MustacheTemplateRegistry()
        )
        self._field_traversal_stack: List[str] = []

    def _should_display_field(self, field_info: FieldInfo) -> bool:
        if isinstance(field_info, SimpleFieldInfo):
            return field_info.default is None
        elif isinstance(field_info, PydanticFieldInfo):
            return cast(PydanticFieldInfo, field_info).default == PydanticUndefined

    @abstractmethod
    def generate(self, field: Field) -> MustacheGenerationResult:
        """Generate the Mustache template for the ``Field``.

        :param field: the ``Field`` to generate a Mustache template for.
        """
        raise NotImplementedError

    def visit_base(
        self,
        base_type: BaseType,
        field_info: FieldInfo,
    ):
        """Visit the base type, only if a default value is not provided for the field.

        :param base_type: the base type.
        :param field_info: information about the field.
        """
        if not self._should_display_field(field_info):
            return

        super().visit_base(base_type, field_info)


class DocumentationGenerator(ABC):
    """Base class of all documentation generators."""

    @abstractmethod
    def generate(self, service: InferenceService) -> MustacheGenerationResult:
        """Generate documentation for the inference service.

        :param service: the inference service.
        """
        raise NotImplementedError


class TypeFormatter(ABC):
    """Format types for display."""

    @classmethod
    def format(cls, base_type: BaseType, field_info: FieldInfo) -> str:
        """Format the base type. This is a base function for ``BaseType``s.

        :param base_type: the base type.
        :param field_info: information about the field.
        """
        if type(base_type) is PrimitiveType:
            return cls.format_primitive(cast(PrimitiveType, base_type), field_info)
        elif type(base_type) is UnknownType:
            return cls.format_unknown(cast(UnknownType, base_type), field_info)
        elif isinstance(base_type, SpecialType):
            return cls.format_special(cast(SpecialType, base_type), field_info)
        elif type(base_type) is CustomObjectType:
            return cls.format_custom_object(
                cast(CustomObjectType, base_type), field_info
            )
        elif type(base_type) is UnionType:
            return cls.format_union(cast(UnionType, base_type), field_info)
        elif type(base_type) is ListType:
            return cls.format_list(cast(ListType, base_type), field_info)
        elif type(base_type) is TupleType:
            return cls.format_tuple(cast(TupleType, base_type), field_info)
        elif type(base_type) is SetType:
            return cls.format_set(cast(SetType, base_type), field_info)
        elif type(base_type) is DictionaryType:
            return cls.format_dictionary(cast(DictionaryType, base_type), field_info)
        elif type(base_type) is EnumType:
            return cls.format_enum(cast(EnumType, base_type), field_info)
        elif type(base_type) is OptionalType:
            return cls.format_optional(cast(OptionalType, base_type), field_info)
        else:
            return "<type>"

    @classmethod
    def format_special(cls, special_type: SpecialType, field_info: FieldInfo) -> str:
        """Format the special type. This is a base function for ``SpecialType``s.

        :param special_type: the special type.
        :param field_info: information about the field.
        """
        if type(special_type) is PydanticHttpUrlType:
            return cls.format_pydantic_http_url(
                cast(PydanticHttpUrlType, special_type), field_info
            )
        elif type(special_type) is MediaType:
            return cls.format_media(cast(MediaType, special_type), field_info)
        else:
            return "<special>"

    @classmethod
    @abstractmethod
    def format_primitive(
        cls, primitive_type: PrimitiveType, field_info: FieldInfo
    ) -> str:
        """Format the primitive type.

        :param primitive_type: the primitive type.
        :param field_info: information about the field.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def format_union(cls, union_type: UnionType, field_info: FieldInfo) -> str:
        """Format the union type.

        :param union_type: the primitive type.
        :param field_info: information about the field.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def format_list(cls, list_type: ListType, field_info: FieldInfo) -> str:
        """Format the list type.

        :param list_type: the list type.
        :param field_info: information about the field.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def format_tuple(cls, tuple_type: TupleType, field_info: FieldInfo) -> str:
        """Format the tuple type.

        :param tuple_type: the tuple type.
        :param field_info: information about the field.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def format_set(cls, set_type: SetType, field_info: FieldInfo) -> str:
        """Format the set type.

        :param set_type: the set type.
        :param field_info: information about the field.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def format_dictionary(cls, dict_type: DictionaryType, field_info: FieldInfo) -> str:
        """Format the dictionary type.

        :param dict_type: the dictionary type.
        :param field_info: information about the field.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def format_enum(cls, enum_type: EnumType, field_info: FieldInfo) -> str:
        """Format the enum type.

        :param enum_type: the enum type.
        :param field_info: information about the field.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def format_unknown(cls, unknown_type: UnknownType, field_info: FieldInfo) -> str:
        """Format the unknown type.

        :param unknown_type: the unknown type.
        :param field_info: information about the field.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def format_optional(cls, optional_type: OptionalType, field_info: FieldInfo) -> str:
        """Format the optional type.

        :param optional_type: the optional type.
        :param field_info: information about the field.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def format_pydantic_http_url(
        cls, pydantic_http_url_type: PydanticHttpUrlType, field_info: FieldInfo
    ) -> str:
        """Format the pydantic HttpUrl type.

        :param pydantic_http_url_type: the pydantic HttpUrl type.
        :param field_info: information about the field.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def format_media(cls, media_type: MediaType, field_info: FieldInfo) -> str:
        """Format the media type.

        :param media_type: the media type.
        :param field_info: information about the field.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def format_custom_object(
        cls, custom_object_type: CustomObjectType, field_info: FieldInfo
    ) -> str:
        """Format the custom object type.

        :param custom_object_type: the custom object type.
        :param field_info: information about the field.
        """
        raise NotImplementedError
